# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test initiate load methods."""
import logging
import os
import pathlib
import shutil
import tempfile
import unittest
from io import StringIO
from unittest.mock import patch
import numpy as np
from pyretis.core.box import create_box
from pyretis.core.path import Path
from pyretis.core.particles import Particles, ParticlesExt
from pyretis.core.pathensemble import PathEnsemble, PathEnsembleExt
from pyretis.core.random_gen import RandomGenerator
from pyretis.core.system import System
from pyretis.core.units import units_from_settings
from pyretis.engines.gromacs import GromacsEngine
from pyretis.engines.internal import Langevin
from pyretis.initiation.initiate_load import (
    clean_path,
    read_path_files,
    read_path_files_ext,
    reorderframes,
    _check_path,
    _do_the_dirty_load_job,
)
from pyretis.inout.common import make_dirs
from pyretis.inout.formats.path import PathExtFile, PathIntFile
from pyretis.inout.formats.order import OrderPathFile
from pyretis.inout.formats.energy import EnergyPathFile
from pyretis.inout.settings import (fill_up_tis_and_retis_settings,
                                    parse_settings_file)
from pyretis.orderparameter import Position, PositionVelocity
from pyretis.setup import create_simulation

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))
CP2K = os.path.join(HERE, 'cp2k', 'mockcp2k.py')
GMX = os.path.join(HERE, 'gromacs', 'mockgmx.py')
MDRUN = os.path.join(HERE, 'gromacs', 'mockmdrun.py')
CP2K_DIR = os.path.join(HERE, 'cp2k', 'cp2k_input')
CP2K_LOAD = os.path.join(HERE, 'cp2k', 'pippo')
GMX_DIR = os.path.join(HERE, 'gromacs', 'gmx_input')
GMX_LOAD = os.path.join(HERE, 'gromacs', 'loader')
GMX_LOAD_TRJ = os.path.join(HERE, 'gromacs', 'load-trj')
INTERNAL_LOAD = os.path.join(HERE, 'internal', 'tis', 'loader')
INTERNAL_LOAD_2 = os.path.join(HERE, 'internal', 'retis-load', 'load-sparse')


def remove_file(filename):
    """Remove a file."""
    if os.path.isfile(filename):
        try:
            os.remove(filename)
        except OSError:
            pass


def set_up_system(order, pos_vel, vpot=None, ekin=None, internal=True):
    """Create a system for testing."""
    pos, vel = pos_vel
    system = System()
    if internal:
        system.particles = Particles(dim=3)
        for posi, veli in zip(pos, vel):
            system.add_particle(posi, vel=veli)
    else:
        system.particles = ParticlesExt(dim=3)
        system.add_particle(pos, vel=vel)
    system.order = order
    system.particles.vpot = vpot
    system.particles.ekin = ekin
    return system


def fill_path(path=None, npoints=20, scale=1.):
    """Fill in data for a arbitrary path."""
    if path is None:
        path = Path()
    for i in range(npoints):
        path.append(
            set_up_system([(-20.0 + 2.5 * i) * scale],
                          [[np.ones(3) * i],
                           [np.ones(3) * i]],
                          vpot=i, ekin=i,
                          internal=True)
        )
    return path


def find_files_in_dir(dirname):
    """Return a list of files in the given directory."""
    return [i.name for i in os.scandir(dirname) if i.is_file()]


def read_traj_order_energy(dirname, path_reader):
    """Read trajectory, order parameter and energy files."""
    traj, order, energy = [], [], []
    traj_file_name = os.path.join(dirname, 'traj.txt')
    with path_reader(traj_file_name, 'r') as trajfile:
        traj = list(next(trajfile.load())['data'])
    order_file_name = os.path.join(dirname, 'order.txt')
    with OrderPathFile(order_file_name, 'r') as orderfile:
        order = list(next(orderfile.load())['data'])
    energy_file_name = os.path.join(dirname, 'energy.txt')
    with EnergyPathFile(energy_file_name, 'r') as energyfile:
        for i in energyfile.load():
            energy = i['data']
            break
    return traj, order, energy


def path_traj_order_energy(path):
    """Get trajectory, order parameters and energies from a path."""
    traj, order, energy = [], [], {'time': [], 'vpot': [], 'ekin': []}
    for i, phasepoint in enumerate(path.phasepoints):
        vel = -1 if phasepoint.particles.get_vel() else 1
        traj.append([i, os.path.basename(phasepoint.particles.get_pos()[0]),
                     phasepoint.particles.get_pos()[1], vel])
        order.append(np.array([i, *phasepoint.order], dtype=float))
        energy['time'].append(i)
        energy['vpot'].append(phasepoint.particles.vpot)
        energy['ekin'].append(phasepoint.particles.ekin)
    for key, val in energy.items():
        energy[key] = np.array(val)
    return traj, order, energy


def path_traj_order_energy_int(path):
    """Get trajectory, order parameters and energies from a path."""
    traj, order, energy = [], [], {'time': [], 'vpot': [], 'ekin': []}
    for i, phasepoint in enumerate(path.phasepoints):
        traj.append({'pos': phasepoint.particles.get_pos(),
                     'vel': phasepoint.particles.get_vel()})
        order.append(np.array([i, *phasepoint.order]))
        energy['time'].append(i)
        energy['vpot'].append(phasepoint.particles.vpot)
        energy['ekin'].append(phasepoint.particles.ekin)
    for key, val in energy.items():
        energy[key] = np.array(val)
    return traj, order, energy


def compare_trajs(path_load, path):
    """Compare loaded trajectory with source."""
    traj, order, energy = read_traj_order_energy(path_load, PathExtFile)
    trajp, orderp, energyp = path_traj_order_energy(path)

    if not (len(order) == len(orderp)
            or len(traj) == len(trajp)):
        return False
    for i, j in zip(order, orderp):
        if not np.allclose(i, j):
            return False
    for traji, trajj in zip(traj, trajp):
        trajjj = [str(i) for i in trajj]
        for i, j in zip(traji, trajjj):
            if not i == j:
                return False
    for term in ['time', 'ekin', 'vpot']:
        if not len(energy[term]) == len(energyp[term]):
            return False
        if not np.allclose(energy[term], energyp[term]):
            return False
    return True


def compare_trajs_int(path_load, path):
    """Compare internal trajectory with source."""
    traj, order, energy = read_traj_order_energy(path_load, PathIntFile)
    trajp, orderp, energyp = path_traj_order_energy_int(path)
    status = True
    if not (len(order) == len(orderp)
            or len(traj) == len(trajp)):
        status = False
    for i, j in zip(order, orderp):
        if not np.allclose(i, j, atol=1e-6):
            status = False
    for traji, trajj in zip(traj, trajp):
        for i in ('pos', 'vel'):
            if not np.allclose(traji[i], trajj[i]):
                status = False
    term = 'time'
    if not len(energy[term]) == len(energyp[term]):
        status = False
    if not np.allclose(energy[term], energyp[term]):
        status = False
    for term in ('ekin', 'vpot'):
        if not len(energy[term]) == len(energyp[term]):
            status = False
        try:
            if not np.allclose(energy[term], energyp[term]):
                status = False
        except TypeError:  # Infinite errors are ok
            pass
    return status


class TestReadPathFilesExt(unittest.TestCase):
    """Run the tests for the reading external trajectories."""

    def test_read_path_ext(self):
        """Test reading of external trajectories."""
        with tempfile.TemporaryDirectory() as tempdir:
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(GMX_LOAD, load_dir)
            path_ensemble = PathEnsembleExt(23, (-0.26, 0.02, 0.2),
                                            exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            for i in path_ensemble.directories():
                make_dirs(i)
            files = find_files_in_dir(path_ensemble.directory['accepted'])
            self.assertEqual(0, len(files))
            gmx = GromacsEngine(GMX, MDRUN, gmx_dir,
                                timestep=0.002,
                                subcycles=3,
                                exe_path=os.path.join(HERE, 'gromacs'),
                                maxwarn=1,
                                gmx_format='gro', write_vel=True,
                                write_force=False)
            gmx.exe_dir = path_ensemble.directory['accepted']
            path = Path(None)
            system = System(box=None)
            system.particles = ParticlesExt()
            ensemble = {'system': system,
                        'order_function': None,
                        'path_ensemble': path_ensemble,
                        'engine': gmx}
            with patch('sys.stdout', new=StringIO()):
                path_ok = read_path_files_ext(path, path_load, ensemble)
                self.assertEqual(path_ok, (True, 'ACC'))
            files = find_files_in_dir(path_ensemble.directory['accepted'])
            self.assertEqual(2, len(files))
            self.assertIn('trajF.trr', files)
            self.assertIn('trajB.trr', files)
            cmpok = compare_trajs(path_load, path)
            self.assertTrue(cmpok)

    def test_read_path_ext_no_order(self):
        """Test reading of external trajectories with missing order.txt."""
        orderparameter = PositionVelocity(1472, dim='x', periodic=True)
        with tempfile.TemporaryDirectory() as tempdir:
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(GMX_LOAD, load_dir)
            path_ensemble = PathEnsembleExt(23, [1.78, 1.795, 1.80],
                                            exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            remove_file(os.path.join(path_load, 'order.txt'))
            for i in path_ensemble.directories():
                make_dirs(i)
            files = find_files_in_dir(path_ensemble.directory['accepted'])
            self.assertEqual(0, len(files))
            files = find_files_in_dir(path_load)
            self.assertEqual(2, len(files))
            self.assertIn('energy.txt', files)
            self.assertIn('traj.txt', files)
            self.assertNotIn('order.txt', files)
            gmx = GromacsEngine(GMX, MDRUN, gmx_dir,
                                timestep=0.002,
                                subcycles=3,
                                exe_path=os.path.join(HERE, 'gromacs'),
                                maxwarn=1,
                                gmx_format='gro', write_vel=True,
                                write_force=False)
            gmx.exe_dir = path_ensemble.directory['accepted']
            path = Path(None)
            system = System(box=None)
            system.particles = ParticlesExt()
            ensemble = {'system': system,
                        'path_ensemble': path_ensemble,
                        'order_function': orderparameter,
                        'engine': gmx}
            with patch('sys.stdout', new=StringIO()):
                path_ok = read_path_files_ext(path, path_load, ensemble)
                self.assertEqual(path_ok, (True, 'ACC'))
            files = find_files_in_dir(path_ensemble.directory['accepted'])
            self.assertEqual(2, len(files))
            self.assertIn('trajF.trr', files)
            self.assertIn('trajB.trr', files)

    def test_read_path_ext_no_energy(self):
        """Test reading of external trajectories with missing order.txt."""
        with tempfile.TemporaryDirectory() as tempdir:
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(GMX_LOAD, load_dir)
            path_ensemble = PathEnsembleExt(23, (-0.26, 0.02, 0.2),
                                            exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            remove_file(os.path.join(path_load, 'energy.txt'))
            for i in path_ensemble.directories():
                make_dirs(i)
            files = find_files_in_dir(path_ensemble.directory['accepted'])
            self.assertEqual(0, len(files))
            files = find_files_in_dir(path_load)
            self.assertEqual(2, len(files))
            self.assertNotIn('energy.txt', files)
            self.assertIn('traj.txt', files)
            self.assertIn('order.txt', files)
            gmx = GromacsEngine(GMX, MDRUN, gmx_dir,
                                timestep=0.002,
                                subcycles=3,
                                exe_path=os.path.join(HERE, 'gromacs'),
                                maxwarn=1,
                                gmx_format='gro', write_vel=True,
                                write_force=False)
            gmx.exe_dir = path_ensemble.directory['accepted']
            path = Path(None)
            system = System(box=None)
            system.particles = ParticlesExt()
            ensemble = {'system': system,
                        'path_ensemble': path_ensemble,
                        'order_function': None,
                        'engine': gmx}
            with patch('sys.stdout', new=StringIO()):
                path_ok = read_path_files_ext(path, path_load, ensemble)
                self.assertEqual(path_ok, (True, 'ACC'))
            files = find_files_in_dir(path_ensemble.directory['accepted'])
            self.assertEqual(2, len(files))
            self.assertIn('trajF.trr', files)
            self.assertIn('trajB.trr', files)
            for i in path.phasepoints:
                self.assertIsNone(i.particles.ekin)
                self.assertIsNone(i.particles.vpot)

    def test_check_path(self):
        """Test checking of the path."""
        with tempfile.TemporaryDirectory() as tempdir:
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(GMX_LOAD, load_dir)
            path_ensemble = PathEnsembleExt(23, [-0.26, 0.02, 0.2],
                                            exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            for i in path_ensemble.directories():
                make_dirs(i)
            gmx = GromacsEngine(GMX, MDRUN, gmx_dir,
                                timestep=0.002,
                                subcycles=3,
                                exe_path=os.path.join(HERE, 'gromacs'),
                                maxwarn=1,
                                gmx_format='gro', write_vel=True,
                                write_force=False)
            gmx.exe_dir = path_ensemble.directory['accepted']
            path = Path(None)
            system = System(box=None)
            system.particles = ParticlesExt()
            ensemble = {'system': system,
                        'path_ensemble': path_ensemble,
                        'order_function': None,
                        'engine': gmx}
            with patch('sys.stdout', new=StringIO()):
                read_path_files_ext(path,
                                    path_load, ensemble)
            accept, status = _check_path(path, path_ensemble)
            self.assertTrue(accept)
            self.assertEqual(status, 'ACC')
            path_ensemble = PathEnsembleExt(23, [-0.26, 0.2, 0.4],
                                            exe_dir=tempdir)
            accept, status = _check_path(path, path_ensemble)
            self.assertFalse(accept)
            self.assertEqual(status, 'EWI')
            path_r = path.reverse(
                Position(0, dim='x', periodic=False)
            )
            path_ensemble = PathEnsembleExt(23, [-0.26, 0.02, 0.2],
                                            exe_dir=tempdir)
            accept, status = _check_path(path_r, path_ensemble)
            self.assertFalse(accept)
            self.assertEqual(status, 'SWI')
            path.append(
                set_up_system([-0.26, 0.0], [('somefile.trr', 0), True],
                              internal=False)
            )
            path_ensemble = PathEnsembleExt(23, [-0.26, 0.3, 0.3],
                                            exe_dir=tempdir)
            accept, status = _check_path(path, path_ensemble)
            self.assertFalse(accept)
            self.assertEqual(status, 'NCR')


class TestReadPathFiles(unittest.TestCase):
    """Run the tests for reading internal trajectories."""

    def test_read_path(self):
        """Test reading of internal trajectories."""
        with tempfile.TemporaryDirectory() as tempdir:
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(INTERNAL_LOAD, load_dir)
            path_ensemble = PathEnsemble(1, [-0.9, -0.8, -0.7],
                                         exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            path = Path(None)
            engine = Langevin(0.002, 0.3)
            box = create_box(periodic=[False])
            system = System(box=box)
            system.particles = Particles()
            ensemble = {'system': system,
                        'path_ensemble': path_ensemble,
                        'order_function': None,
                        'engine': engine}
            with patch('sys.stdout', new=StringIO()):
                accept, status = read_path_files(path, path_load, ensemble)
                self.assertTrue(accept)
                self.assertEqual(status, 'ACC')
            cmpt = compare_trajs_int(path_load, path)
            self.assertTrue(cmpt)

    def test_read_path_noorder(self):
        """Test reading of internal trajectories with missing order.txt."""
        orderparameter = PositionVelocity(0, dim='x', periodic=False)
        with tempfile.TemporaryDirectory() as tempdir:
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(INTERNAL_LOAD, load_dir)
            path_ensemble = PathEnsemble(1, [-0.9, -0.8, -0.7],
                                         exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            remove_file(os.path.join(path_load, 'order.txt'))
            path = Path(None)
            engine = Langevin(0.002, 0.3)
            box = create_box(periodic=[False])
            system = System(box=box)
            system.particles = Particles(dim=1)
            ensemble = {'system': system,
                        'path_ensemble': path_ensemble,
                        'order_function': orderparameter,
                        'engine': engine}
            with patch('sys.stdout', new=StringIO()):
                accept, status = read_path_files(path, path_load, ensemble)
                self.assertTrue(accept)
                self.assertEqual(status, 'ACC')
            shutil.copyfile(
                os.path.join(INTERNAL_LOAD, '001', 'order.txt'),
                os.path.join(path_load, 'order.txt')
            )
            cmpt = compare_trajs_int(path_load, path)
            self.assertTrue(cmpt)

    def test_read_path_noenergy(self):
        """Test reading of internal trajectories with missing energy.txt."""
        with tempfile.TemporaryDirectory() as tempdir:
            load_dir = os.path.join(tempdir, 'load')
            shutil.copytree(INTERNAL_LOAD, load_dir)
            path_ensemble = PathEnsemble(1, [-0.9, -0.8, -0.7],
                                         exe_dir=tempdir)
            path_load = os.path.join(
                load_dir, path_ensemble.ensemble_name_simple)
            remove_file(os.path.join(path_load, 'energy.txt'))
            path = Path(None)
            engine = Langevin(0.002, 0.3)
            box = create_box(periodic=[False])
            system = System(box=box)
            system.particles = Particles(dim=1)
            ensemble = {'system': system,
                        'order_function': None,
                        'path_ensemble': path_ensemble,
                        'engine': engine}
            with patch('sys.stdout', new=StringIO()):
                accept, status = read_path_files(path, path_load, ensemble)
                self.assertTrue(accept)
                self.assertEqual(status, 'ACC')
            for i in path.phasepoints:
                self.assertIsNone(i.particles.ekin)
                self.assertIsNone(i.particles.vpot)


class TestInitiateLoad(unittest.TestCase):
    """Test the full initiate load for a TIS simulation."""

    @staticmethod
    def _set_up_for_load_int(settings, tempdir):
        """Set up a system for testing the initiate load method."""
        settings['simulation']['exe_path'] = os.path.join(tempdir)
        load_dir = os.path.join(tempdir, 'loader')

        shutil.copytree(os.path.join(HERE, 'internal', 'tis'), tempdir)

        settings['initial-path']['load_folder'] = load_dir
        settings['engine']['exe_path'] = tempdir
        settings['engine']['input_path'] = tempdir

        del settings['ensemble']
        fill_up_tis_and_retis_settings(settings)

        units_from_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            simulation = create_simulation(settings)
        simulation.set_up_output(settings)
        return simulation, load_dir

    @staticmethod
    def _test_dirs(tempdir):
        """Test the creation of the data structure."""
        for test1 in ['000', '001', '002', '003']:
            dirt = os.path.join(tempdir, test1)
            if not os.path.exists(dirt):
                return False
            for test2 in ['traj', 'generate', 'accepted']:
                if not os.path.exists(os.path.join(dirt, test2)):
                    return False
            for test3 in ['order.txt', 'energy.txt', 'pathensemble.txt']:
                if not os.path.exists(os.path.join(tempdir, test1, test3)):
                    return False
        return True

    @staticmethod
    def _set_up_for_load(settings, tempdir, copy_load=True):
        """Set up a system for testing the initiate load method."""
        settings['simulation']['exe_path'] = tempdir
        load_dir = os.path.join(
            tempdir, settings['initial-path']['load_folder']
        )
        settings['initial-path']['load_folder'] = load_dir
        ens_name = ['000', '001', '002', '003', '004']
        for i in range(len(settings['ensemble'])):
            os.makedirs(os.path.join(tempdir, ens_name[i]))
            settings['ensemble'][i]['simulation']['exe_path'] = tempdir
            settings['ensemble'][i]['initial-path']['load_folder'] =\
                load_dir
            settings['ensemble'][i]['engine']['exe_path'] =\
                os.path.join(HERE, 'internal', 'tis')
            settings['ensemble'][i]['engine']['input_path'] =\
                os.path.join(HERE, 'internal', 'tis')
        if copy_load:
            shutil.copytree(INTERNAL_LOAD, load_dir)
        shutil.copyfile(
            os.path.join(HERE, 'internal', 'tis', 'initial.xyz'),
            os.path.join(tempdir, 'initial.xyz')
        )

        units_from_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            simulation = create_simulation(settings)
        simulation.set_up_output(settings)
        return simulation

    @staticmethod
    def _set_up_for_load_2(settings, tempdir, copy_load=True):
        """Set up a second system for testing the initiate load method."""
        settings['simulation']['exe_path'] = tempdir
        load_dir = os.path.join(
            tempdir, settings['initial-path']['load_folder'])

        settings['engine']['gmx'] = GMX
        settings['engine']['mdrun'] = MDRUN
        settings['engine']['exe_path'] = tempdir
        settings['engine']['input_path'] = tempdir
        settings['simulation']['exe_path'] = tempdir
        settings['initial-path']['load_folder'] = load_dir

        del settings['ensemble']
        fill_up_tis_and_retis_settings(settings)

        if copy_load:
            shutil.copytree(INTERNAL_LOAD_2, load_dir)
        units_from_settings(settings)
        simulation = create_simulation(settings)
        simulation.set_up_output(settings)
        return simulation

    def test_internal_load(self):
        """Test the load initialise for internal trajectories."""
        settings = parse_settings_file(os.path.join(HERE, 'internal', 'tis',
                                                    'tis.rst'))

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'internal', 'tis')
            with patch('sys.stdout', new=StringIO()):
                simulation, _ = \
                    self._set_up_for_load_int(settings, tempdir)
                init = simulation.initiate(settings)
                self.assertTrue(init)
            # We expect now that we have loaded for 001:
            self.assertEqual(len(simulation.ensembles), 1)
            self.assertEqual(simulation.ensembles[0]['path_ensemble']
                             .ensemble_number, 1)
            # Check that the path contains what it should:
            path = simulation.ensembles[0]['path_ensemble'].last_path
            # Check that we created an ensemble dir
            ensemble_dir = os.path.join(tempdir, '001')
            self.assertTrue(os.path.isdir(ensemble_dir))
            # All memory is dumped to file.
            self.assertTrue(compare_trajs_int(ensemble_dir, path))
            # Just delete simulation to force closing of files:
            del simulation
            self.assertTrue(compare_trajs_int(ensemble_dir, path))

    def test_fail_wrong_engine(self):
        """Test that the test fail for an unsupported engine type."""
        settings = parse_settings_file(os.path.join(HERE, 'internal', 'tis',
                                                    'tis.rst'))
        with tempfile.TemporaryDirectory() as tempdir:
            simulation = self._set_up_for_load(settings, tempdir)
            ensemble = simulation.ensembles[0]
            ensemble['engine'].engine_type = 'this-should-fail!'
            with patch('sys.stdout', new=StringIO()):
                with self.assertRaises(ValueError):
                    simulation.initiate(settings)

    def test_internal_load_sparse(self):
        """Test the load initialise for internal trajectories."""
        settings = parse_settings_file(os.path.join(HERE, 'internal',
                                                    'retis-load',
                                                    'reload.rst'))
        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'internal', 'retis-load')

            with patch('sys.stdout', new=StringIO()):
                simulation = self._set_up_for_load_2(settings, tempdir)
                init = simulation.initiate(settings)
                self.assertTrue(init)

            paths = []
            for i in range(3):
                paths.append(
                    simulation.ensembles[i]['path_ensemble'].last_path)
            del simulation
            self.assertTrue(compare_trajs_int(os.path.join(tempdir, '000'),
                                              paths[0]))

            self.assertTrue(os.path.exists(os.path.join(tempdir, '000')))
            self.assertTrue(os.path.exists(os.path.join(tempdir, '001',
                                                        'traj')))
            self.assertTrue(os.path.exists(os.path.join(tempdir, '002',
                                                        'generate')))
            self.assertTrue(os.path.exists(os.path.join(tempdir, '001',
                                                        'pathensemble.txt')))
            self.assertTrue(os.path.exists(os.path.join(tempdir, '002',
                                                        'order.txt')))
            self.assertEqual(paths[0].status, 'ACC')
            for i in range(1, 2):
                self.assertEqual(paths[i].status, 'ACC')

    def test_internal_explorer(self):
        """Test the explorer for internal trajectories."""
        settings = parse_settings_file(os.path.join(HERE, 'internal',
                                                    'retis-load',
                                                    'reload.rst'))
        settings['simulation']['task'] = 'explore'
        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'internal', 'retis-load')

            with patch('sys.stdout', new=StringIO()):
                simulation = self._set_up_for_load_2(settings, tempdir)
                init = simulation.initiate(settings)
                self.assertTrue(init)

                # Does it work too?
                res = simulation.step()
                self.assertEqual(res['status-1'], 'EXP')
                self.assertEqual(res['status-2'], 'EXP')

    def test_internal_explorer_bad_init(self):
        """Test the explorer for internal trajectories."""
        settings = parse_settings_file(os.path.join(HERE, 'internal',
                                                    'retis-load',
                                                    'reload.rst'))
        settings['simulation']['task'] = 'explore'
        # A case where the initial configurations do not have frames after the
        # ensemble interface.
        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'internal', 'retis-load')
            with patch('sys.stdout', new=StringIO()):
                settings['tis']['maxlength'] = 100
                settings['simulation']['interfaces'][0] = -2
                settings['simulation']['interfaces'].append(1000)
                simulation = self._set_up_for_load_2(settings, tempdir)
                init = simulation.initiate(settings)
                self.assertTrue(init)

                # Does it work too?
                res = simulation.step()
                self.assertEqual(res['status-0'], 'EXP')
                # It shall try anyhow.
                self.assertEqual(res['status-2'], 'EXP')

    def test_internal_load_sparse_and_kick(self):
        """Test the load initialise that fixes paths."""
        settings = parse_settings_file(os.path.join(HERE, 'internal',
                                                    'retis-load',
                                                    'reload.rst'))
        settings['simulation']['interfaces'] = [-1.2, -0.4, -0.2]
        # Load poor initial paths
        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'internal', 'retis-load')

            with patch('sys.stdout', new=StringIO()):
                simulation = self._set_up_for_load_2(settings, tempdir)
                # Check that we fail first
                with self.assertRaises(ValueError) as cm:
                    simulation.initiate(settings)
                self.assertIn("load_and_kick", str(cm.exception))

        settings = parse_settings_file(os.path.join(HERE, 'internal',
                                                    'retis-load',
                                                    'reload.rst'))
        settings['initial-path']['load_and_kick'] = True
        del settings['ensemble']
        settings['simulation']['interfaces'] = [-1.2, -0.4, -0.2]
        fill_up_tis_and_retis_settings(settings)

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'internal', 'retis-load')

            with patch('sys.stdout', new=StringIO()):
                simulation2 = self._set_up_for_load_2(settings, tempdir)
                init = simulation2.initiate(settings)
                self.assertTrue(init)

    def test_initiate_load_fail(self):
        """Test that the initiate load fails for non-existing load folder."""
        settings = parse_settings_file(
            os.path.join(HERE, 'internal', 'tis', 'tis.rst')
        )
        with tempfile.TemporaryDirectory() as tempdir:
            with patch('sys.stdout', new=StringIO()):
                simulation = self._set_up_for_load(settings, tempdir,
                                                   copy_load=False)
                with self.assertRaises(FileNotFoundError):
                    simulation.initiate(settings)

    def test_initiate_load_fail2(self):
        """Test that the initiate load fails for insufficient file inputs."""
        settings = parse_settings_file(os.path.join(HERE, 'internal',
                                                    'retis-load',
                                                    'reload.rst'))
        with tempfile.TemporaryDirectory() as tempdir:
            with patch('sys.stdout', new=StringIO()):
                simulation = self._set_up_for_load_2(settings, tempdir)
            # Instead to change the inputs, we change the requirements.
            simulation.ensembles[0]['path_ensemble'].interfaces = [1,
                                                                   100,
                                                                   100000]
            with patch('sys.stdout', new=StringIO()):
                with self.assertRaises(ValueError) as err:
                    _ = simulation.initiate(settings)
                self.assertTrue('000 does not satisfy the relative ensemble'
                                in str(err.exception))

    def test_reorderframes(self):
        """Test the reorderframes function."""
        interfaces = [float('-inf'), -10., 10]
        path_ensemble = PathEnsemble(0, interfaces)
        path = fill_path()

        left, _, right = path_ensemble.interfaces
        path_ensemble.start_condition = 'R'

        # Test that the re-order function generated path
        # satisfies the ensemble requirements (R, R case)
        path, _ = reorderframes(path, path_ensemble)
        self.assertTrue(path.ordermin[0] < interfaces[1])
        self.assertTrue(path.ordermax[0] > interfaces[1])
        self.assertEqual(path.get_start_point(left, right), 'R')
        self.assertEqual(path.get_end_point(left, right), 'R')

        # Test that the re-order function generated path
        # satisfies the ensemble requirements (L, L case)
        path_ensemble = PathEnsemble(1, [-18., -10., 60.])
        left, _, right = path_ensemble.interfaces
        path_ensemble.start_condition = 'L'
        path = fill_path()
        path, _ = reorderframes(path, path_ensemble)

        self.assertTrue(path.get_start_point(left, right) == 'L')
        self.assertTrue(path.get_end_point(left, right) == 'L')
        self.assertTrue(path.ordermin[0] < interfaces[1])
        self.assertTrue(path.ordermax[0] > interfaces[1])

        # Test that the re-order function generated path
        # satisfies the ensemble requirements (L, R case)
        path_ensemble = PathEnsemble(2, [-10., 1., 2.])
        left, _, right = path_ensemble.interfaces
        path_ensemble.start_condition = 'L'
        path = fill_path()
        path, _ = reorderframes(path, path_ensemble)

        self.assertTrue(path.get_start_point(left, right) == 'L')
        self.assertFalse(path.get_end_point(left, right) == 'L')

    def test_clean_path(self):
        """Test the clean_path function."""
        interfaces = [-14., -10., 10]
        path_ensemble = PathEnsemble(0, interfaces)

        path = fill_path()
        clean_path(path, path_ensemble)
        self.assertEqual(path.length, 6)
        self.assertTrue(_check_path(path, path_ensemble))

        # Same results?
        clean_path(path, path_ensemble)
        self.assertEqual(path.length, 6)
        self.assertTrue(_check_path(path, path_ensemble))

        # Some intuitive numerical checks: only crossing points
        # shall remain (4 for the 0 interface and one for the B
        # interface).
        path = fill_path(scale=10000)
        interfaces = [-16., 0., 16]
        path_ensemble = PathEnsemble(0, interfaces)
        initial_length = path.length
        path, _ = reorderframes(path, path_ensemble)

        # Numerical checks
        counter_r, counter_l, counter_m = 0, 0, 0
        for _, phasepoint in enumerate(path.phasepoints):
            if phasepoint.order[0] > interfaces[2]:
                counter_r += 1
            if phasepoint.order[0] < interfaces[0]:
                counter_l += 1
            if interfaces[0] < phasepoint.order[0] < interfaces[2]:
                counter_m += 1

        self.assertEqual(path.length, (counter_r + counter_l + counter_m))
        self.assertEqual((counter_r, counter_m, counter_l), (11, 1, 8))
        self.assertTrue(path.length == initial_length)
        self.assertEqual(path.length, 20)
        self.assertTrue(_check_path(path, path_ensemble))

        # This creates a bad path, invalid start but some ok frames
        path = fill_path(scale=0.1)
        interfaces = [-16., 0., 16]
        path_ensemble = PathEnsemble(0, interfaces)
        initial_length = path.length
        path, _ = reorderframes(path, path_ensemble)
        clean_path(path, path_ensemble)

        # Numerical checks
        counter_r, counter_l, counter_m = 0, 0, 0
        for phasepoint in path.phasepoints:
            if phasepoint.order[0] >= interfaces[2]:
                counter_r += 1
            if phasepoint.order[0] < interfaces[0]:
                counter_l += 1
            if interfaces[0] <= phasepoint.order[0] < interfaces[2]:
                counter_m += 1

        self.assertEqual(path.length, (counter_r + counter_l + counter_m))
        self.assertEqual((counter_r, counter_m, counter_l), (0, 21, 0))
        self.assertTrue(path.length > initial_length)
        self.assertEqual(path.length, 21)
        self.assertTrue(_check_path(path, path_ensemble))

        # This creates a good path, bookkeeping case for 0 ens
        path = fill_path(scale=0.1)
        interfaces = [-1., 0., 16]
        path_ensemble = PathEnsemble(0, interfaces)
        initial_length = path.length
        path, _ = reorderframes(path, path_ensemble)
        clean_path(path, path_ensemble)

        # Numerical checks
        counter_r, counter_l, counter_m = 0, 0, 0
        for phasepoint in path.phasepoints:
            if phasepoint.order[0] >= interfaces[2]:
                counter_r += 1
            if phasepoint.order[0] < interfaces[0]:
                counter_l += 1
            if interfaces[0] <= phasepoint.order[0] < interfaces[2]:
                counter_m += 1

        self.assertEqual(path.length, (counter_r + counter_l + counter_m))
        self.assertEqual((counter_r, counter_m, counter_l), (0, 14, 0))
        self.assertTrue(path.length < initial_length)
        self.assertEqual(path.length, 14)
        self.assertTrue(_check_path(path, path_ensemble))

    def test_clean_path_2(self):
        """Test the clean_path function."""
        interfaces = [float('-inf'), -10., -5]
        path_ensemble = PathEnsemble(0, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)

        fill_path(path, npoints=20000, scale=0.01)
        clean_path(path, path_ensemble)
        self.assertEqual(path.length, 5)
        self.assertTrue(_check_path(path, path_ensemble))

        interfaces = [-14., -10., 10]
        path_ensemble = PathEnsemble(0, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)

        fill_path(path)
        clean_path(path, path_ensemble)
        self.assertEqual(path.length, 6)
        self.assertTrue(_check_path(path, path_ensemble))

        # Check that it doesn't fix the impossible.
        # Too large interfaces, all should be kept
        interfaces = [-14000., 0., 1000]
        path_ensemble = PathEnsemble(0, interfaces)

        path = fill_path(npoints=1000, scale=0.16)
        path, _ = reorderframes(path, path_ensemble)
        self.assertEqual((_check_path(path, path_ensemble)),
                         (False, 'EWI'))
        clean_path(path, path_ensemble)
        self.assertEqual(path.length, 21)

        # Too large orderp, almost nothing should remain
        interfaces = [-16., 0., 16]
        path_ensemble = PathEnsemble(0, interfaces)
        path = fill_path(npoints=1000, scale=16016)
        path, _ = reorderframes(path, path_ensemble)
        clean_path(path, path_ensemble)
        self.assertEqual((_check_path(path, path_ensemble)), (False, 'EWI'))
        self.assertEqual(path.length, 5)

    def test_clean_path3(self):
        """Test the clean_path function."""
        interfaces = [float('-inf'), -1., 5]
        path_ensemble = PathEnsemble(0, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=20, scale=0.01)
        path_ensemble.start_condition = ('L', 'R')
        clean_path(path, path_ensemble)
        self.assertEqual(path.length, 20)
        self.assertTrue(_check_path(path, path_ensemble))

        interfaces = [-1., 0., 1]
        path_ensemble = PathEnsemble(0, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=16, scale=0.1)
        path_ensemble.start_condition = ('L', 'R')
        self.assertTrue(path.phasepoints[0].order[0] <= interfaces[0])
        self.assertTrue(path.phasepoints[path.ordermax[1]].order[0] >
                        interfaces[-1])
        clean_path(path, path_ensemble)
        self.assertTrue(path.phasepoints[-1].order[0] <= interfaces[0])
        self.assertEqual(path.length, 11)

        interfaces = [-1., 0., 10]
        path_ensemble = PathEnsemble(0, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=160, scale=0.01)
        path_ensemble.start_condition = ('L', 'R')
        clean_path(path, path_ensemble)
        self.assertFalse(path.phasepoints[0].order[0] <= interfaces[0])
        self.assertFalse(path.phasepoints[-1].order[0] > interfaces[-1])

    def test_clean_reorder_path_repptis(self):
        """Test clean path and reorderframes for pptis ensembles"""
        # 1) body ensemble [i^+-] with i > 0
        # 1A) too long LMR path --> Trim down to LMR
        interfaces = [-1., 0, 1.]
        path_ensemble = PathEnsemble(3, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=20, scale=0.055)
        path_ensemble.start_condition = ('L', 'R')
        clean_path(path, path_ensemble, simtype='repptis')
        self.assertEqual(path.length, 17)
        self.assertTrue(_check_path(path, path_ensemble))
        self.assertTrue(path.phasepoints[0].order[0] < interfaces[0])
        self.assertTrue(path.phasepoints[-1].order[0] > interfaces[-1])
        # 1B) LM* path --> Makes an LML path
        interfaces = [-1., 0, 2.]
        path_ensemble = PathEnsemble(3, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=20, scale=0.055)
        path_ensemble.start_condition = ('L', 'R')
        clean_path(path, path_ensemble, simtype='repptis')
        pathnew, (accept, status) = reorderframes(path, path_ensemble)
        self.assertTrue(accept)
        self.assertEqual(status, 'ACC')
        self.assertTrue(pathnew.phasepoints[0].order[0] < interfaces[0])
        self.assertTrue(pathnew.phasepoints[-1].order[0] < interfaces[0])
        self.assertTrue(pathnew.ordermax[0] > interfaces[1])
        self.assertTrue(_check_path(pathnew, path_ensemble))
        self.assertFalse(_check_path(path, path_ensemble)[0])
        # 2) ensemble [0^+-'] should only give LMR, RML, or LML paths
        interfaces = [-1, -1, 0]
        path_ensemble = PathEnsemble(1, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=20, scale=0.055)
        path_ensemble.start_condition = ('L', 'R')
        clean_path(path, path_ensemble, simtype='repptis')
        self.assertTrue(path.phasepoints[0].order[0] < interfaces[0])
        self.assertTrue(path.phasepoints[-1].order[0] > interfaces[-1])
        self.assertTrue(_check_path(path, path_ensemble))
        # 2b) what if start_condition is only L
        interfaces = [-1, -1, 0]
        path_ensemble = PathEnsemble(1, interfaces)
        rgen = RandomGenerator(seed=0)
        path = Path(rgen, maxlen=1000)
        fill_path(path, npoints=20, scale=0.055)
        path_ensemble.start_condition = 'L'
        clean_path(path, path_ensemble, simtype='repptis')
        self.assertTrue(path.phasepoints[0].order[0] < interfaces[0])
        self.assertTrue(path.phasepoints[-1].order[0] < interfaces[0])
        self.assertTrue(_check_path(path, path_ensemble))

    def test_external_load(self):
        """Test the load initialise for external trajectories."""
        settings = parse_settings_file(
            os.path.join(HERE, 'gromacs', 'tis.rst')
        )
        settings['engine']['gmx'] = GMX
        settings['engine']['mdrun'] = MDRUN

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'gromacs')
            settings['simulation']['exe_path'] = tempdir
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            load_dir = os.path.join(tempdir, 'loader')
            shutil.copytree(GMX_LOAD, load_dir)
            settings['engine']['input_path'] = gmx_dir
            units_from_settings(settings)
            settings['initial-path']['load_folder'] = load_dir
            settings['engine']['exe_path'] = tempdir

            with patch('sys.stdout', new=StringIO()):
                # Reset inputs for this example set up in the tempdir
                del settings['ensemble']
                fill_up_tis_and_retis_settings(settings)
                # Calculate the WF path weight:
                settings['ensemble'][0]['tis']['high_accept'] = True
                settings['ensemble'][0]['tis']['shooting_move'] = 'wf'
                simulation = create_simulation(settings)
                simulation.set_up_output(settings)
                init = simulation.initiate(settings)
                self.assertTrue(init)
            for to_test in ['accepted', 'order.txt',
                            'energy.txt', 'pathensemble.txt']:
                self.assertTrue(os.path.exists(os.path.join(tempdir, '023',
                                                            to_test)))
            self.assertEqual(simulation.ensembles[0]['path_ensemble']
                             .last_path.status, 'ACC')
            self.assertEqual(simulation.ensembles[0]['path_ensemble']
                             .last_path.weight, 42.0)

    def test_external_load_cp2k(self):
        """Test the load initialise for external trajectories with cp2k."""
        settings = parse_settings_file(
            os.path.join(HERE, 'cp2k', 'retis.rst'))
        settings['engine']['cp2k'] = CP2K

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'cp2k')
            cp2k_dir = os.path.join(tempdir, 'cp2k_input')
            shutil.copytree(CP2K_DIR, cp2k_dir)
            load_dir = os.path.join(tempdir, 'pippo')
            shutil.copytree(CP2K_LOAD, load_dir)
            settings['engine']['input_path'] = cp2k_dir
            units_from_settings(settings)
            settings['initial-path']['load_folder'] = load_dir
            settings['simulation']['exe_path'] = tempdir
            settings['engine']['exe_path'] = tempdir
            with patch('sys.stdout', new=StringIO()):
                # Reset inputs for this example set up in the tempdir
                del settings['ensemble']
                fill_up_tis_and_retis_settings(settings)
                for i in ['000', '001', '002', '003']:
                    edir = os.path.join(load_dir, i)
                    _do_the_dirty_load_job(load_dir, edir)
                simulation = create_simulation(settings)
            simulation.set_up_output(settings)
            with patch('sys.stdout', new=StringIO()):
                init = simulation.initiate(settings)
                self.assertTrue(init)
            self.assertTrue(self._test_dirs(tempdir))

            for i in range(3):
                self.assertEqual(
                    simulation.ensembles[i]['path_ensemble'].last_path.status,
                    'ACC')
                path = simulation.ensembles[i]['path_ensemble'].last_path
                for j in range(int(path.length/3)):
                    point1 = path.phasepoints[j]
                    point2 = path.phasepoints[-1-j]
                    self.assertEqual(point1.order[0], point2.order[0])

    def test_external_load_external(self):
        """Test the load initialise for external trajectories."""
        settings = parse_settings_file(
            os.path.join(HERE, 'gromacs', 'auto-load.rst'))
        orderp_file = os.path.join(HERE, 'gromacs', 'orderp.py')
        settings['engine']['gmx'] = GMX
        settings['engine']['mdrun'] = MDRUN
        top_file = settings['initial-path']['top_file']

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'gromacs')
            settings['simulation']['exe_path'] = tempdir
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            shutil.copy(orderp_file, tempdir)
            tmp_top_file = os.path.join(tempdir, top_file)
            load_dir = os.path.join(tempdir, 'loader')
            shutil.copytree(GMX_LOAD, load_dir)
            settings['engine']['input_path'] = gmx_dir

            units_from_settings(settings)
            settings['initial-path']['load_folder'] = load_dir
            settings['initial-path']['top_file'] = tmp_top_file
            settings['engine']['exe_path'] = tempdir

            # Reset inputs for this example set up in the tempdir
            del settings['ensemble']
            fill_up_tis_and_retis_settings(settings)

            with patch('sys.stdout', new=StringIO()):
                simulation = create_simulation(settings)
                simulation.set_up_output(settings)
                init = simulation.initiate(settings)
                self.assertTrue(init)
            self.assertTrue(self._test_dirs(tempdir))

            for i in range(4):
                self.assertEqual(
                    simulation.ensembles[i]
                    ['path_ensemble'].last_path.status, 'ACC')

            self.assertTrue(os.path.exists(os.path.join(load_dir, '004',
                                                        'traj.txt')))

            for i in range(1, 4):
                path = simulation.ensembles[i]['path_ensemble'].last_path
                for j in range(int(path.length / 2) - 1):
                    point1 = path.phasepoints[j]
                    point2 = path.phasepoints[j + 1]
                    self.assertTrue(point1.order[0] <= point2.order[0])

    def test_external_load_external_2(self):
        """Test the load initialise for external trajectories."""
        settings = parse_settings_file(
            os.path.join(HERE, 'gromacs', 'auto-load.rst'))
        orderp_file = os.path.join(HERE, 'gromacs', 'orderp.py')
        settings['simulation']['interfaces'] = [2.78, 2.88, 3.01, 3.80, 14.00]
        settings['engine']['gmx'] = GMX
        settings['engine']['mdrun'] = MDRUN
        top_file = settings['initial-path']['top_file']

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'gromacs')
            settings['simulation']['exe_path'] = tempdir
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            shutil.copy(orderp_file, tempdir)
            tmp_top_file = os.path.join(tempdir, top_file)
            shutil.copy(os.path.join(HERE, 'gromacs', top_file), tmp_top_file)
            load_dir = os.path.join(tempdir, 'loader')
            shutil.copytree(GMX_LOAD, load_dir)
            settings['engine']['input_path'] = gmx_dir
            settings['initial-path']['load_folder'] = load_dir
            units_from_settings(settings)

            # Reset inputs for this example set up in the tempdir
            del settings['ensemble']
            fill_up_tis_and_retis_settings(settings)

            with patch('sys.stdout', new=StringIO()):
                simulation = create_simulation(settings)
                simulation.set_up_output(settings)
                init = simulation.initiate(settings)
                self.assertTrue(init)
            self.assertTrue(self._test_dirs(tempdir))

            for i in range(3):
                self.assertEqual(
                    simulation.ensembles[i]['path_ensemble'].last_path.status,
                    'ACC')

            self.assertTrue(os.path.exists(os.path.join(load_dir, '004',
                                                        'traj.txt')))

            for i in range(1, 3):
                path = simulation.ensembles[i]['path_ensemble'].last_path
                for j in range(int(path.length/3)):
                    self.assertEqual(path.phasepoints[j].order[0],
                                     path.phasepoints[-1-j].order[0])

    def test_external_load_external_trr(self):
        """Test the load initialise for external trajectories."""
        settings = parse_settings_file(
            os.path.join(HERE, 'gromacs', 'auto-load.rst')
        )
        orderp_file = os.path.join(HERE, 'gromacs', 'orderp.py')
        settings['engine']['gmx'] = GMX
        settings['engine']['mdrun'] = MDRUN
        del settings['initial-path']['top_file']

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'gromacs')
            settings['simulation']['exe_path'] = tempdir
            gmx_dir = os.path.join(tempdir, 'gmx_input')
            shutil.copytree(GMX_DIR, gmx_dir)
            shutil.copy(orderp_file, tempdir)
            load_dir = os.path.join(tempdir, 'load-trj')
            shutil.copytree(GMX_LOAD_TRJ, load_dir)
            settings['engine']['input_path'] = gmx_dir
            settings['engine']['exe_path'] = tempdir

            units_from_settings(settings)
            settings['initial-path']['load_folder'] = load_dir

            # Reset inputs for this example set up in the tempdir
            del settings['ensemble']
            fill_up_tis_and_retis_settings(settings)

            with patch('sys.stdout', new=StringIO()):
                simulation = create_simulation(settings)
                simulation.set_up_output(settings)
                init = simulation.initiate(settings)
                self.assertTrue(init)
            self.assertTrue(self._test_dirs(tempdir))

            for i in range(4):
                self.assertEqual(
                    simulation.ensembles[i]['path_ensemble'].last_path.status,
                    'ACC')

            self.assertTrue(os.path.exists(os.path.join(load_dir, '004',
                                                        'traj.txt')))

            exp_len = [None, 11, 11, 11]
            for i in range(1, 4):
                path = simulation.ensembles[i]['path_ensemble'].last_path
                self.assertEqual(path.length, exp_len[i])

    @staticmethod
    def _create_empty_gro_files(tempdir, nfiles):
        """Create a given number of empty .gro files."""
        grofiles = []
        for i in range(nfiles):
            gro = f'conf{i:04d}.gro'
            grofile = os.path.join(tempdir, gro)
            pathlib.Path(grofile).touch()
            grofiles.append(gro)
        return grofiles

    def test_load_job(self):
        """Test some aspects of the dirty load method."""
        settings = parse_settings_file(os.path.join(HERE, 'internal', 'tis',
                                                    'tis.rst'))

        with tempfile.TemporaryDirectory() as tempdirbase:
            tempdir = os.path.join(tempdirbase, 'gromacs')
            self._set_up_for_load(settings, tempdir)
            # Remove the ensemble directory to test what happens:
            edir = os.path.join(tempdir, '001')
            self.assertTrue(os.path.isdir(tempdirbase))
            self.assertTrue(os.path.isdir(tempdir))
            self.assertTrue(os.path.isdir(edir))
            shutil.rmtree(edir)
            self.assertFalse(os.path.isdir(edir))
            # And remove the initial.xyz:
            remove_file(os.path.join(tempdir, 'initial.xyz'))
            with self.assertRaises(FileNotFoundError):
                _do_the_dirty_load_job(tempdir, edir)
            # Create some fake .gro files to see what happens next:
            grofiles = self._create_empty_gro_files(tempdir, 3)
            accepted = os.path.join(edir, 'accepted')
            self.assertFalse(os.path.isdir(edir))
            self.assertFalse(os.path.isdir(accepted))
            _do_the_dirty_load_job(tempdir, edir)
            # Check that edir was created:
            self.assertTrue(os.path.isdir(edir))
            # Check that the accepted dir was crated:
            self.assertTrue(os.path.isdir(accepted))
            # And that the gro files were copied:
            for i in grofiles:
                self.assertTrue(os.path.isfile(os.path.join(accepted, i)))
                remove_file(os.path.join(tempdir, i))
            # Remove the accepted dir and see what happens:
            shutil.rmtree(accepted)
            with self.assertRaises(FileNotFoundError):
                _do_the_dirty_load_job(tempdir, edir)
            # Add some files in tempdir and see what happens:
            grofiles = self._create_empty_gro_files(tempdir, 3)
            self.assertFalse(os.path.isdir(accepted))
            _do_the_dirty_load_job(tempdir, edir)
            self.assertTrue(os.path.isdir(accepted))
            # Check that files were copied:
            for i in grofiles:
                self.assertTrue(os.path.isfile(os.path.join(accepted, i)))
                self.assertTrue(os.path.isfile(os.path.join(tempdir, i)))
                remove_file(os.path.join(tempdir, i))
            shutil.rmtree(accepted)
            # Add some files in edir and see what happens:
            grofiles = self._create_empty_gro_files(edir, 3)
            self.assertFalse(os.path.isdir(accepted))
            _do_the_dirty_load_job(tempdir, edir)
            self.assertTrue(os.path.isdir(accepted))
            # Check that files were moved:
            for i in grofiles:
                self.assertTrue(os.path.isfile(os.path.join(accepted, i)))
                self.assertFalse(os.path.isfile(os.path.join(tempdir, i)))


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This is a mock engine for external testing.

This script is inteneded to be used as an external engine and it
will simply just create the output files.
"""
import sys
import os
import numpy as np
from pyretis.inout.formats.xyz import write_xyz_trajectory


def write_energy(outfile, steps=10):
    """Store CP2K style energies."""
    header = ' '.join(['# Step Nr.', 'Time[fs]', 'Kin.[a.u.]',
                       'Temp[K]', 'Pot.[a.u.]', 'Cons Qty[a.u.]',
                       'UsedTime[s]'])
    fmt = '{:10d} {:8.6f} {:8.6f} {:8.6f} {:8.6f} {:8.6f} {:8.6f}'
    with open(outfile, 'w') as output:
        output.write(f'{header}\n')
        for i in range(steps):
            time = i * 0.5
            kin = i * 0.1
            temp = i * 100
            pot = -0.1 * i
            txt = fmt.format(i, time, kin, temp, pot, i, i)
            output.write(f'{txt}\n')


def write_wfn_file(outfile):
    """Write some gibberish data to represent a wfn file."""
    if os.path.isfile(outfile):
        # Just keep one backup file. This is useful for the testing.
        backup = f'{outfile}.bak-1'
        os.rename(outfile, backup)
    with open(outfile, 'w') as output:
        output.write('Some gibberish')


def write_xyz_file(outfile, factor, steps=10):
    """Write some mock xyz-data."""
    names = ['H', 'H']
    box = [1., 2., 3.]
    fac = np.array([factor, np.ones(3) + factor])
    for i in range(steps):
        pos = np.ones((2, 3)) * i * fac
        vel = np.zeros((2, 3))
        write_xyz_trajectory(outfile, pos, vel, names, box, step=i,
                             append=True)


def read_project_name(filename):
    """Get the CP2K project name."""
    with open(filename, 'r') as infile:
        for lines in infile:
            if lines.find('PROJECT') != -1:
                projname = lines.split()[-1]
                return projname
    return None


def write_cp2k_restart(filename, factorx, factorv, steps=10):
    """Write a CP2K-like restart file."""
    xyz = (steps - 1) * factorx
    vel = (steps - 1) * factorv
    if os.path.isfile(filename):
        # Just keep one backup file. This is useful for the testing.
        backup = f'{filename}.bak-1'
        os.rename(filename, backup)
    with open(filename, 'w') as outfile:
        outfile.write('&FORCE_EVAL\n')
        outfile.write('&SUBSYS\n')
        outfile.write('&COORD\n')
        for i in range(2):
            xyzi = xyz + np.ones(3) * i
            outfile.write(
                '{:s} {:8.6f} {:8.6f} {:8.6f}\n'.format('H', *xyzi)
            )
        outfile.write('&END COORD\n')
        outfile.write('&CELL\n')
        outfile.write(
            '{:s} {:8.6f} {:8.6f} {:8.6f}\n'.format('A', 1., 0., 0.)
        )
        outfile.write(
            '{:s} {:8.6f} {:8.6f} {:8.6f}\n'.format('B', 0., 2., 0.)
        )
        outfile.write(
            '{:s} {:8.6f} {:8.6f} {:8.6f}\n'.format('C', 0., 0., 3.)
        )
        outfile.write('&END CELL\n')
        outfile.write('&VELOCITY\n')
        for i in range(2):
            veli = vel + np.ones(3) * i
            outfile.write(
                '{:8.6f} {:8.6f} {:8.6f}\n'.format(*veli)
            )
        outfile.write('&END VELOCITY\n')
        outfile.write('&END SUBSYS\n')
        outfile.write('&END FORCE_EVAL\n')


if __name__ == '__main__':
    # pylint: disable=invalid-name
    print('This is the mock CP2K command')
    args = sys.argv
    name = read_project_name(args[-1])
    if not name:
        sys.exit(1)

    efile = f'{name}-1.ener'
    write_energy(efile, steps=10)

    wfile = f'{name}-RESTART.wfn'
    write_wfn_file(wfile)

    xfile = f'{name}-pos-1.xyz'
    xfac = np.array([0.1, 0.2, 0.3])
    write_xyz_file(xfile, xfac, steps=10)

    vfile = f'{name}-vel-1.xyz'
    vfac = np.array([1.1, 1.2, 1.3])
    write_xyz_file(vfile, vfac, steps=10)

    rfile = f'{name}-1.restart'
    write_cp2k_restart(rfile, xfac, vfac, steps=10)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test a single step with the CP2K external engine."""
import os
import time
import colorama
from pyretis.core import System, create_box, ParticlesExt
from pyretis.inout.common import make_dirs
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.engines.cp2k import (
    CP2KEngine,
)


def clean_dir(dirname):
    """Remove ALL files in the given directory!"""
    for files in os.listdir(dirname):
        filename = os.path.join(dirname, files)
        if os.path.isfile(filename):
            os.remove(filename)


def run_step(engine, system, exe_dir='forward-single-step'):
    """Run the engine forward in time, in steps.

    Parameters
    ----------
    engine : object like :py:class:`.ExternalMDEngine
        Engine to use for propagation.
    system : object like :py:class:`.System`
        The system we are propagation.
    exe_dir : string
        The folder to use for the execution.
    """
    print_to_screen(f'\nRunning a single CP2K step in "{exe_dir}"',
                    level='message')
    make_dirs(exe_dir)
    folder = os.path.abspath(exe_dir)
    clean_dir(folder)
    engine.exe_dir = folder
    engine.step(system, 'single')
    print_to_screen('Propagation done!')


def test_genvel(engine, input_file, exe_dir='genvel'):
    """Test generation of velocities.

    Parameters
    ----------
    engine : object like :py:class:`.ExternalMDEngine
        Engine to use for the generation.
    input_file : string
        The input configuration used for CP2K.
    exe_dir : string, optional
        The directory where we will be running CP2K.
    """
    print_to_screen(f'\nRunning CP2K genvel step in "{exe_dir}"',
                    level='message')
    make_dirs(exe_dir)
    folder = os.path.abspath(exe_dir)
    clean_dir(folder)
    engine.exe_dir = folder
    engine._prepare_shooting_point(input_file)
    print_to_screen('Propagation done!')


def main():
    """Execute the test."""
    settings = parse_settings_file('engine.rst')
    engine_settings = settings['engine']
    engine = CP2KEngine(
        engine_settings['cp2k'],
        engine_settings['input_path'],
        engine_settings['timestep'],
        engine_settings['subcycles'],
        engine_settings.get('extra_files', [])
    )
    print_to_screen(f'Testing engine: {engine}', level='info')
    print_to_screen(f'Time step: {engine.timestep}')
    print_to_screen(f'Subcycles: {engine.subcycles}')
    system = System(units='gromacs',
                    box=create_box(cell=[100, 100, 100]),
                    temperature=500)
    system.particles = ParticlesExt(dim=3)
    initial_conf = engine.input_files['conf']
    system.particles.set_pos((initial_conf, None))
    system.particles.set_vel(False)

    start = time.perf_counter()
    run_step(engine, system, exe_dir='forward-single-step')
    end = time.perf_counter()
    print_to_screen(f'Time spent: {end - start}', level='info')

    start = time.perf_counter()
    test_genvel(engine, 'cp2k_input/initial.xyz', exe_dir='genvel')
    end = time.perf_counter()
    print_to_screen(f'Time spent: {end - start}', level='info')


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This is a simple RETIS example.

Here we do nothing fancy, we will just get to know
some of the objects in PyRETIS.

Have fun!
"""
import colorama
import numpy as np
from pyretis.inout.settings import (fill_up_tis_and_retis_settings,
                                    _add_default_settings,
                                    _add_specific_default_settings)
from pyretis.core import System, create_box, Particles
from pyretis.initiation import initiate_path_simulation
from pyretis.setup import (
    create_force_field,
    create_orderparameter,
    create_simulation
)

# Let us define the simulation:
SETTINGS = {}
# Basic settings for the simulation:
SETTINGS['simulation'] = {
    'task': 'retis',
    'steps': 10,
    'interfaces': [-0.9, -0.8, -0.7, -0.6, -0.5, -0.4, -0.3, 1.0],
}
# Basic settings for the system:
SETTINGS['system'] = {'units': 'reduced', 'temperature': 0.07, 'dimensions': 1}
# Basic settings for the Langevin integrator:
SETTINGS['engine'] = {
    'class': 'Langevin',
    'gamma': 0.3,
    'high_friction': False,
    'seed': 0,
    'timestep': 0.002
}
# Potential parameters:
# The potential is: `V_\text{pot} = a x^4 - b (x - c)^2`
SETTINGS['potential'] = [
    {
        'class': 'DoubleWell',
        'a': 1.0, 'b': 2.0, 'c': 0.0,
    }
]
# Settings for the order parameter:
SETTINGS['orderparameter'] = {
    'class': 'Position',
    'dim': 'x',
    'index': 0,
    'periodic': False
}
# TIS specific settings:
SETTINGS['tis'] = {
    'freq': 0.5,
    'maxlength': 20000,
    'aimless': True,
    'allowmaxlength': False,
    'sigma_v': -1,
    'seed': 0,
    'zero_momentum': False,
    'rescale_energy': False
}
SETTINGS['initial-path'] = {'method': 'kick'}
# RETIS specific settings:
SETTINGS['retis'] = {
    'swapfreq': 0.5,
    'relative_shoots': None,
    'nullmoves': True,
    'swapsimul': True
}


def set_up_system(settings):
    """Set up the system.

    Parameters
    ----------
    settings : dict
        The settings required to set up the system.

    Returns
    -------
    sys : object like :py:class:`.System`
        A system object we can use in a simulation.

    """
    box = create_box(periodic=[False])
    sys = System(temperature=settings['system']['temperature'],
                 units=settings['system']['units'], box=box)
    sys.forcefield = create_force_field(settings)
    sys.order_function = create_orderparameter(settings)
    sys.particles = Particles(dim=1)
    sys.add_particle(np.array([-1.0]), mass=1, name='Ar', ptype=0)
    return sys


def print_step_results(ensembles, retis_result):
    """Print out RETIS results.

    Parameters
    ----------
    ensembles : list of dicts. They contain:
        *  PathEnsemble: objects like :py:class:`.PathEnsemble`
           The different path ensembles we are simulating.
    retis_result : list of lists
        The results of a RETIS simulation step.

    """
    for ensemble in ensembles:
        path_ensemble = ensemble['path_ensemble']
        name = path_ensemble.ensemble_name
        idx = path_ensemble.ensemble_number
        print(f'Move in {name}')
        accepted = retis_result[f'accept-{idx}']
        name_of_move = retis_result[f'move-{idx}']
        print(f'\tType: {name_of_move}')
        if name_of_move == 'swap':
            idx2 = retis_result[f'all-{idx}']['swap-with']
            name2 = ensembles[idx2]['path_ensemble'].ensemble_name
            print(f'\tSwapping: {name2} -> {name}')
        elif name_of_move == 'tis':
            trial_path = retis_result[f'path-{idx}']
            if trial_path.generated[0] == 'sh':
                tis_move = 'shooting'
            elif trial_path.generated[0] == 'tr':
                tis_move = 'time-reversal'
            else:
                tis_move = 'unknown'
            print(f'\tTIS move: {tis_move}')
        print(f'\tResult: {accepted}')


def main():
    """Just run the simulation."""
    colorama.init(autoreset=True)
    _add_default_settings(SETTINGS)
    _add_specific_default_settings(SETTINGS)
    SETTINGS['system']['obj'] = set_up_system(SETTINGS)
    fill_up_tis_and_retis_settings(SETTINGS)
    simulation = create_simulation(SETTINGS)
    print(simulation)
    print('# INITIATING TRAJECTORIES...')

    ensembles = simulation.ensembles
    for i, _ in enumerate(initiate_path_simulation(simulation, SETTINGS)):
        path_ensemble = ensembles[i]['path_ensemble']
        name = path_ensemble.ensemble_name
        print(f'Info about ensemble {name}:')
        print(path_ensemble)
        print('Info about the initial path:')
        print(path_ensemble.last_path)
        print('')
    # We can interact directly with points in trajectories,
    # here is a simple example
    path = ensembles[2]['path_ensemble'].last_path
    first = True
    for i, point in enumerate(path.phasepoints):
        order = point.order[0]
        pos = point.particles.pos
        vel = point.particles.vel
        if order > -0.8 and first:
            print('First crossing of -0.8 for [1^+]:')
            print(f'\tStep: {i}')
            print(f'\tlambda: {order}')
            print(f'\tPosition and velocity: {pos}  {vel}')
            first = False

    # Let us do one more step:
    print('Running a single RETIS step...')
    result = simulation.step()
    for ensemble in ensembles:
        path_ensemble = ensemble['path_ensemble']
        name = path_ensemble.ensemble_name
        idx = path_ensemble.ensemble_number
        print(f'Move in {name}')
        status = result[f'status-{idx}']
        accepted = result[f'accept-{idx}']
        name_of_move = result[f'move-{idx}']
        # `status` is equal to "ACC" if the move is accepted
        # otherwise it will be one of:
        # 'MCR': 'Momenta change rejection',
        # 'BWI': 'Backward trajectory end at wrong interface',
        # 'BTL': 'Backward trajectory too long (detailed balance condition)',
        # 'BTX': 'Backward trajectory too long (max-path exceeded)',
        # 'KOB': 'Kicked outside of boundaries',
        # 'FTL': 'Forward trajectory too long (detailed balance condition)',
        # 'FTX': 'Forward trajectory too long (max-path exceeded)',
        # 'NCR': 'No crossing with middle interface'
        print(f'\tType: {name_of_move}')
        if name_of_move == 'swap':
            print(status, accepted)
            # If this is the case, the result is on the form
            # [move, accepted, .., swap-with] where swap-with is the
            # ensemble we are trying to swap with.
            idx2 = result[f'all-{idx}']['swap-with']
            name2 = ensembles[idx2]['path_ensemble'].ensemble_name
            print(f'\tSwapping: {name2} -> {name}')
        elif name_of_move == 'tis':
            trial_path = result[f'path-{idx}']
            if trial_path.generated[0] == 'sh':
                tis_move = 'shooting'
            elif trial_path.generated[0] == 'tr':
                tis_move = 'time-reversal'
            else:
                tis_move = 'unknown'
            print(f'\tTIS move: {tis_move}')
        print(f'\tResult: {status}')

    # Run the rest of the simulation:
    while not simulation.is_finished():
        result = simulation.step()
        print(f"Simulation step: {result['cycle']['step']}")
        print_step_results(ensembles, result)
        print('')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A custom engine for Leap Frog integration."""
from pyretis.engines import MDEngine


class LeapFrog(MDEngine):
    """Perform Leap Frog integration."""

    def __init__(self, timestep):
        """Initiate the engine.

        Parameters
        ----------
        timestep : float
            Set the time step for the integrator.
        """
        super().__init__(timestep, 'Leap Frog integrator', dynamics='NVE')
        self.half_timestep = self.timestep * 0.5
        self.half_timestep_sq = 0.5 * self.timestep**2

    def integration_step(self, system):
        """Perform one step for the Leap Frog integrator."""
        particles = system.particles
        imass = particles.imass

        # get current acceleration:
        acc_t = particles.force * imass
        # update positions:
        particles.pos += (self.timestep * particles.vel +
                          self.half_timestep_sq * acc_t)
        # update forces:
        system.potential_and_force()
        # update acceleration:
        acc_t2 = particles.force * imass
        # update velocities:
        particles.vel += self.half_timestep * (acc_t + acc_t2)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of using a Lennard-Jones potential implemented in C."""
import logging
import os
import sys
import numpy as np
from pyretis.forcefield.potentials import PairLennardJonesCut
from pyretis.forcefield.potentials.pairpotentials import (
    generate_pair_interactions
)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
try:
    import ljc
except ImportError as exc:
    MSG = ('Could not import external C library.'
           '\nPlease compile with: "python setup.py build_ext --inplace"')
    raise ImportError(MSG) from exc


__all__ = ['PairLennardJonesCutC']


class PairLennardJonesCutC(PairLennardJonesCut):
    r"""class PairLennardJonesCutC(PairLennardJonesCut).

    This class implements as simple Lennard-Jones 6-12 potential which
    employs a simple cut-off and can be shifted. The potential energy
    (:math:`V_\text{pot}`) is defined in the usual way for an
    interacting pair of particles a distance :math:`r` apart,

    .. math::

       V_\text{pot} = 4 \varepsilon \left( x^{12} - x^{6} \right),

    where :math:`x = \sigma/r` and :math:`\varepsilon`
    and :math:`\sigma` are the potential parameters. The parameters are
    stored as attributes of the potential and we store one set for each
    kind of pair interaction. Parameters can be generated with a
    specific mixing rule by the force field.

    Attributes
    ----------
    params : dict
        The parameters for the potential. This dict is assumed to
        contain parameters for pairs, i.e. for interactions.
    _lj : dict of numpy.array
        [1] Lennard-Jones parameters used for calculation of the force.
        Keys are the pairs (particle types) that may interact.
        Calculated as: ``48.0 * epsilon * sigma**12``
        [2] Lennard-Jones parameters used for calculation of the force.
        Keys are the pairs (particle types) that may interact.
        Calculated as: ``24.0 * epsilon * sigma**6``
        [3] Lennard-Jones parameters used for calculation of the potential.
        Keys are the pairs (particle types) that may interact.
        Calculated as: ``4.0 * epsilon * sigma**12``
        [4] Lennard-Jones parameters used for calculation of the potential.
        Keys are the pairs (particle types) that may interact.
        Calculated as: ``4.0 * epsilon * sigma**6``
    _offset : numpy.array
        Potential values for shifting the potential if requested.
        This is the potential evaluated at the cutoff.
    _rcut2 : numpy.array
        The squared cut-off for each interaction type.
        Keys are the pairs (particle types) that may interact.

    """

    def __init__(self, dim=3, shift=True, mixing='geometric',
                 desc='Lennard-Jones pair potential (C)'):
        """Initialise the Lennard-Jones potential.

        Parameters
        ----------
        dim : int
            The dimensionality to use.
        shift : boolean
            Determines if the potential should be shifted or not.
        mixing : string
            Determines how we should mix potential parameters.

        """
        super().__init__(dim=dim, desc=desc, mixing=mixing)
        self.shift = shift
        self.ntype = 0

    def set_parameters(self, parameters):
        """Update all parameters.

        Here, we generate pair interactions, since that is what this
        potential actually is using.

        Parameters
        ----------
        parameters : dict
            The input base parameters.

        """
        self.params = {}
        pair_param = generate_pair_interactions(parameters, self.mixing)
        self.ntype = max(int(np.sqrt(len(pair_param))), 2)
        self._lj = {'1': np.zeros((self.ntype, self.ntype)),
                    '2': np.zeros((self.ntype, self.ntype)),
                    '3': np.zeros((self.ntype, self.ntype)),
                    '4': np.zeros((self.ntype, self.ntype))}
        self._rcut2 = np.zeros_like(self._lj[1])
        self._offset = np.zeros_like(self._lj[1])
        for pair, value in pair_param.items():
            eps_ij = value['epsilon']
            sig_ij = value['sigma']
            rcut = value['rcut']
            self._lj[1][pair] = 48.0 * eps_ij * sig_ij**12
            self._lj[2][pair] = 24.0 * eps_ij * sig_ij**6
            self._lj[3][pair] = 4.0 * eps_ij * sig_ij**12
            self._lj[4][pair] = 4.0 * eps_ij * sig_ij**6
            self._rcut2[pair] = rcut**2
            vcut = 0.0
            if self.shift and not np.isclose(rcut, 0):
                vcut = 4.0 * eps_ij * ((sig_ij / rcut)**12 -
                                       (sig_ij / rcut)**6)
            self._offset[pair] = vcut
            self.params[pair] = pair_param[pair]

    def potential(self, system):
        """Calculate the potential energy for the Lennard-Jones interaction.

        Parameters
        ----------
        system : object like `System` from `pyretis.core.system`
            The system we are evaluating the potential in.

        Returns
        -------
        v_pot : float
            The potential energy.

        """
        particles = system.particles
        box = system.box
        v_pot = ljc.potential(
            particles.pos,
            box.length,
            box.ilength,
            self._lj,
            self._offset,
            self._rcut2,
            particles.ptype,
            particles.npart,
            box.dim,
            self.ntype
        )
        return v_pot

    def force(self, system):
        """Calculate the force for the Lennard-Jones interaction.

        We also calculate the virial here, since the force
        is evaluated.

        Parameters
        ----------
        system : object like `System` from `pyretis.core.system`
            The system we are evaluating the force in.

        Returns
        -------
        forces : numpy.array
            The forces on the particles.
        virial : numpy.array
            The virial obtained from the forces.

        """
        particles = system.particles
        box = system.box
        forces = np.zeros((particles.npart, box.dim))
        virial = np.zeros((box.dim, box.dim))
        ljc.force(
            particles.pos,
            box.length,
            box.ilength,
            self._lj,
            self._rcut2,
            particles.ptype,
            forces,
            virial,
            particles.npart,
            box.dim,
            self.ntype
        )
        return forces, virial

    def potential_and_force(self, system):
        """Calculate potential and force for the Lennard-Jones interaction.

        Since the force is evaluated, the virial is also calculated.

        Parameters
        ----------
        system : object like `System` from `pyretis.core.system`
            The system we are evaluating the potential and force in.

        Note
        ----
        Currently, the virial is only calculated for all the particles.
        It is not calculated as a virial per atom. The virial
        per atom might be useful to obtain a local pressure or stress,
        however, this needs some consideration. Perhaps it's best to
        fully implement this as a method of planes or something similar.

        Returns
        -------
        out[0] : float
            The potential energy as a float.
        out[1] : numpy.array
            The force as a numpy.array of the same shape as the
            positions in `particles.pos`.
        out[2] : numpy.array
            The virial, as a symmetric matrix with dimensions
            (dim, dim) where dim is given by the box/system dimensions.

        """
        particles = system.particles
        box = system.box
        forces = np.zeros((particles.npart, box.dim))
        virial = np.zeros((box.dim, box.dim))
        vpot = ljc.force_and_pot(
            particles.pos,
            box.length,
            box.ilength,
            self._lj,
            self._offset,
            self._rcut2,
            particles.ptype,
            forces,
            virial,
            particles.npart,
            box.dim,
            self.ntype
        )
        return vpot, forces, virial



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for formatting order parameter data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OrderFormatter (:py:class:`.OrderFormatter`)
    A class for formatting order parameter data.

OrderPathFormatter (:py:class:`.OrderPathFormatter`)
    A class for formatting order parameter data for paths.

OrderFile (:py:class:`.OrderFile`)
    A class for handling PyRETIS order parameter files.

OrderPathFile (:py:class:`.OrderPathFile`)
    A class for handling PyRETIS order parameter files for paths.

"""
import logging
import numpy as np
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.fileio import FileIO, read_some_lines
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = [
    'OrderFormatter',
    'OrderPathFormatter',
    'OrderFile',
    'OrderPathFile',
]


class OrderFormatter(OutputFormatter):
    """A class for formatting order parameter data.

    The format for the order file is column-based and the columns are:

    1) Time.

    2) Main order parameter.

    3) Collective variable 1

    4) Collective variable 2

    5) ...

    """

    # Format for order files. Note that we don't know how many parameters
    # we need to format yet.
    ORDER_FMT = ['{:>10d}', '{:>12.6f}']

    def __init__(self, name='OrderFormatter'):
        """Initialise a `OrderFormatter` formatter."""
        header = {'labels': ['Time', 'Orderp'], 'width': [10, 12]}
        super().__init__(name, header=header)

    def format_data(self, step, orderdata):
        """Format order parameter data.

        Parameters
        ----------
        step : int
            This is the current step number.
        orderdata : list of floats
            These are the order parameters.

        Yields
        ------
        out : string
            The strings to be written.

        """
        towrite = [self.ORDER_FMT[0].format(step)]
        for orderp in orderdata:
            towrite.append(self.ORDER_FMT[1].format(orderp))
        out = ' '.join(towrite)
        return out

    def format(self, step, data):
        """Yield formatted order parameters. See :py:meth:`.format_data`."""
        yield self.format_data(step, data)

    def load(self, filename):
        """Read order parameter data from a file.

        Since this class defines how the data is formatted it is also
        convenient to have methods for reading the data defined here.
        This method will read entire blocks of data from a file into
        memory. This will be slow for large files and this method
        could be converted to also yield the individual "rows" of
        the blocks, rather than the full blocks themselves.

        Parameters
        ----------
        filename : string
            The path/file name of the file we want to open.

        Yields
        ------
        data_dict : dict
            This is the order parameter data in the file.

        See Also
        --------
        :py:func:`.read_some_lines`.

        """
        for blocks in read_some_lines(filename, self.parse):
            data_dict = {'comment': blocks['comment'],
                         'data': np.array(blocks['data'])}
            yield data_dict


class OrderPathFormatter(OrderFormatter):
    """A class for formatting order parameter data for paths."""

    def __init__(self):
        """Initialise."""
        super().__init__(name='OrderPathFormatter')
        self.print_header = False

    def format(self, step, data):
        """Format the order parameter data from a path.

        Parameters
        ----------
        step : int
            The cycle number we are creating output for.
        data : tuple or list
            Here, data[0] contains an object
            like :py:class:`.PathBase` which is the path we are
            creating output for data[1] contains the status for
            this path.

        Yields
        ------
        out : string
            The strings to be written.

        """
        path, status = data[0], data[1]
        if not path:  # E.g. when null-moves are False.
            return
        move = path.generated
        yield f'# Cycle: {step}, status: {status}, move: {move}'
        yield self.header
        for i, phasepoint in enumerate(path.phasepoints):
            yield self.format_data(i, phasepoint.order)


class OrderFile(FileIO):
    """A class for handling PyRETIS order parameter files."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the order file with correct formatter."""
        super().__init__(filename, file_mode, OrderFormatter(), backup=backup)


class OrderPathFile(FileIO):
    """A class for handling PyRETIS order parameter path files."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the order path file with correct formatter."""
        super().__init__(filename, file_mode, OrderPathFormatter(),
                         backup=backup)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Remove the container class from <div> items.

This is just to remove some unwanted containers in the HTML
output.

"""
from sphinx.writers import html


class PyretisTranslator(html.HTMLTranslator):
    """Create a translator that removes the container class."""

    def visit_container(self, node):
        self.body.append(self.starttag(node, 'div', CLASS=''))


def setup(app):
    app.set_translator('html', PyretisTranslator)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for path ensemble analysis."""
import copy
import logging
import os
import unittest
import numpy as np
from pyretis.analysis.path_analysis import (
    analyse_path_ensemble,
    analyse_repptis_ensemble,
    match_probabilities,
    perm_calculations,
    retis_flux,
    retis_rate,
    skip_paths
)
from pyretis.core.path import Path
from pyretis.core.system import System
from pyretis.inout.analysisio.analysisio import (
    repptis_running_pcross_analysis,
)
from pyretis.inout.formats.pathensemble import PathEnsembleFile
from pyretis.inout.settings import SECTIONS
from .help import turn_on_logging
logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))


def add_some_paths(ensemble, raw_data):
    """Add some fake path to the given ensemble."""
    interfaces = ensemble.interfaces
    for _, path in enumerate(raw_data.get_paths()):
        dummy_path = Path(None, maxlen=None, time_origin=0)
        dummy_path.generated = path['generated']
        # Add fake points for the interfaces:
        start, _, end = path['interface']
        order = ([((path['ordermax'][0] + path['ordermin'][0]) * 0.5, 0)] *
                 path['length'])
        for i in (start, end):
            if i == 'L':
                order.append([interfaces[0] - 0.1])
            elif start == 'R':
                order.append([interfaces[2] + 0.1])
            else:
                order.append([0.5*(interfaces[0] + interfaces[2])])
        order[path['ordermax'][1]] = path['ordermax']
        order[path['ordermin'][1]] = path['ordermin']
        for i in order:
            phasepoint = System()
            phasepoint.order = i
            dummy_path.append(phasepoint)
        ensemble.add_path_data(dummy_path, path['status'])


class AnalysePathEnsembleTest(unittest.TestCase):
    """Test that we run the analysis for PathEnsemble results."""

    def test_no_accept_ens_1(self):
        """Test when no accepted paths are available in 0+."""
        ens = {
            'ensemble_number': 1,
            'interfaces': [-float('inf'), -0.9, -0.9],
            'file': 'pathensemble_no_acc.txt',
            'test': {'fluxlength': (0, 1541.7987987987988)},
            }
        settings = {'analysis': SECTIONS['analysis']}
        settings['tis'] = {'detect': -0.9}
        filename = os.path.join(HERE, ens['file'])
        raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
        with self.assertRaises(AssertionError) as err:
            analyse_path_ensemble(raw_data, settings)
        self.assertIn("1", str(err.exception))

    def test_no_accept_ens_0(self):
        """Test when no accepted paths are available in 0-."""
        ens = {
            'ensemble_number': 0,
            'interfaces': [-float('inf'), -0.9, -0.9],
            'file': 'pathensemble_no_acc.txt',
            'test': {'fluxlength': (0, 1541.7987987987988)},
            }
        settings = {'analysis': SECTIONS['analysis']}
        settings['tis'] = {'detect': -0.9}
        filename = os.path.join(HERE, ens['file'])
        raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
        with self.assertRaises(AssertionError) as err:
            analyse_path_ensemble(raw_data, settings)
        self.assertIn("0", str(err.exception))

    def test_path_analysis(self):
        """Test the path ensemble analysis."""
        ensembles = [
            {
                'ensemble_number': 0,
                'interfaces': [-float('inf'), -0.9, -0.9],
                'file': 'pathensemble000.txt',
                'test': {'fluxlength': (0, 1541.7987987987988)},
            },
            {
                'ensemble_number': 1,
                'interfaces': [-0.9, -0.9, 1],
                'file': 'pathensemble001.txt',
                'test': {'prun': (-1, 223/996)},
            },
            {
                'ensemble_number': 2,
                'interfaces': [-0.9, -0.8, 1],
                'file': 'pathensemble002.txt',
                'test': {'prun': (-1, 145/996)},
            },
            {
                'ensemble_number': 3,
                'interfaces': [-0.9, -0.7, 1],
                'file': 'pathensemble003.txt',
                'test': {'prun': (-1, 125/999)},
            },
            {
                'ensemble_number': 4,
                'interfaces': [-0.9, -0.71, 1],
                'file': 'pathensemble004.txt',
                'test': {'prun': (-1, 125/966)},
            },
        ]
        settings = {'analysis': SECTIONS['analysis']}
        results = []
        detects = [-0.9, -0.8, -0.7, -0.6, -0.6]
        raw_data = None
        val, key = None, None
        for ens, dec in zip(ensembles, detects):
            filename = os.path.join(HERE, ens['file'])
            settings['tis'] = {'detect': dec}
            raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
            res = analyse_path_ensemble(raw_data, settings)
            for key, val in ens['test'].items():
                self.assertAlmostEqual(val[-1], res[key][val[0]])
            results.append(analyse_path_ensemble(raw_data, settings))
        match = match_probabilities(results[1:], detects[1:], settings)
        self.assertAlmostEqual(match['prob'], 0.0005277540804156307)
        flux = retis_flux(results[0], results[1], 0.002)
        self.assertAlmostEqual(flux[0], 0.26572079970779655)
        self.assertAlmostEqual(flux[1], 0.02422284635774688)
        rate = retis_rate(match['prob'], match['relerror'], flux[0], flux[1])
        self.assertAlmostEqual(rate[0], 0.00014023523629709417)
        self.assertAlmostEqual(rate[1], 0.4734317724627789)

        # Re-check the last one:
        settings['analysis']['maxblock'] = 0
        res = analyse_path_ensemble(raw_data, settings)
        self.assertAlmostEqual(val[-1], res[key][val[0]])

    def test_skipping(self):
        """Test the skip_lines funcionality."""
        ens = {
            'ensemble_number': 0,
            'interfaces': [-float('inf'), -0.1, -0.1],
            'file': 'permeability/pathensemble.txt',
            }
        settings = {'analysis': SECTIONS['analysis']}
        # Skip more trajectories than there are
        # Make a copy to prevent overwrite
        settings = copy.deepcopy(settings)
        settings['tis'] = {'detect': -0.1}
        settings['analysis']['skip'] = 10000
        filename = os.path.join(HERE, ens['file'])
        raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
        with self.assertRaisesRegex(AssertionError, "No accepted paths"):
            analyse_path_ensemble(raw_data, settings)
        # Make sure we can skip all trajectories
        settings['analysis']['skip'] = 49
        _ = analyse_path_ensemble(raw_data, settings)
        # Test more skipping than possible (Hard to hit normally)
        with self.assertRaisesRegex(RuntimeError, "Skipping more trajs than"):
            skip_paths([1, 1], 2, 3)
        # Test exact skipping of all
        weights, nacc = skip_paths([1, 2], 3, 3)
        assert weights == [0, 0]
        assert nacc == 1

        # See if this also works in ensemble 1
        ens['ensemble_number'] = 1
        raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
        settings['analysis']['skip'] = 10000
        with self.assertRaisesRegex(AssertionError, "No accepted paths"):
            analyse_path_ensemble(raw_data, settings)

    def test_permeability_path_analysis_interfaces(self):
        """Test the permeability calculations."""
        ens = {
            'ensemble_number': 0,
            'interfaces': [-float('inf'), -0.1, -0.1],
            'file': 'permeability/pathensemble.txt',
            }

        settings = dict(analysis=SECTIONS['analysis'],
                        simulation={'permeability': True})

        settings['analysis']['tau_ref_bin'] = [-0.175, -0.125]
        settings['tis'] = {'detect': -0.1}
        filename = os.path.join(HERE, ens['file'])
        raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
        # Test no zero left
        result = analyse_path_ensemble(raw_data, settings)
        # Check if zero left is not overwritten
        assert result['interfaces'][0] == -float('inf')
        # Check if hr is properly calculated
        assert result['xi'][-1] == 30/49
        # The error is about 3%
        assert 0.03 < result['xierror'][4] < 0.04
        # Check tau/dz
        assert result['tau'][-1] == (8/49)/(-0.125+0.175)
        # Tau error is about 19%
        assert 0.19 < result['tauerror'][4] < 0.20

        # Test with zero left
        settings['simulation']['zero_left'] = -0.2
        result = analyse_path_ensemble(raw_data, settings)

        # Assert zero left overwritten and no data change
        assert result['interfaces'][0] == -0.2
        # Check if hr is properly calculated
        assert result['xi'][-1] == 30/49
        # The error is about 3%
        assert 0.03 < result['xierror'][4] < 0.04
        # Check tau/dz
        assert result['tau'][-1] == (8/49)/(-0.125+0.175)
        # Tau error is about 19%
        assert 0.19 < result['tauerror'][4] < 0.20

        # Now we are going to use the miss aligned order.txt (cycle 0 => 10)
        # It should not matter for the result as it is part of the init
        op_dir = filename.rsplit('/', 1)[0]
        correct = (op_dir+'/order.txt', op_dir+'/order_correct.txt')
        mis = (op_dir+'/order.txt', op_dir+'/order_misaligned.txt')
        os.rename(*correct)
        os.rename(*mis[::-1])
        raw_data2 = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
        with turn_on_logging():
            with self.assertLogs('pyretis.analysis.path_analysis',
                                 level=logging.WARNING):
                result2 = analyse_path_ensemble(raw_data2, settings)

        # Move files back
        os.rename(*mis)
        os.rename(*correct[::-1])

        assert (result['xi'] == result2['xi']).all()
        assert (result['tau'] == result2['tau']).all()

        # Now assert the other calculations
        result0 = {'out': result}
        # Make_fake_cros
        pcross = 0.5
        perr = 0.5
        xi, tau, perm, xi_err, tau_err, perm_err = perm_calculations(
            result0, pcross, perr)

        assert xi == result['xi'][-1]
        assert tau == result['tau'][-1]
        assert xi_err == result['xierror'][4]
        assert tau_err == result['tauerror'][4]
        # Check permeation formula and error propagation
        assert perm == result['xi'][-1]*(1/result['tau'][-1])*pcross
        assert perm_err == (xi_err**2+tau_err**2+perr**2)**(1/2)

    def test_path_analysis_pptis(self):
        """Test the path ensemble analysis for (RE)PPTIS simulations."""
        pptisdir = "pptis"
        ensembles = [
            {
                'ensemble_number': 0,
                'interfaces': [0.1, 0.2, 0.2],
                'file': os.path.join(pptisdir, 'pathensemble0.txt'),
                'test': {'fluxlength': (0, 91.98)},
            },
            {
                'ensemble_number': 1,
                'interfaces': [0.2, 0.2, 0.325],
                'file': os.path.join(pptisdir, 'pathensemble1.txt'),
                'test': {'prun_sl': (-1, 0.07526881720430108),
                         'prun_sr': (-1, 1.)},
            },
            {
                'ensemble_number': 2,
                'interfaces': [0.2, 0.325, 0.55],
                'file': os.path.join(pptisdir, 'pathensemble2.txt'),
                'test': {'prun_sl': (-1, 0.010638297872340425),
                         'prun_sr': (-1, 1.)},
            },
            {
                'ensemble_number': 3,
                'interfaces': [0.325, 0.55, 0.69],
                'file': os.path.join(pptisdir, 'pathensemble3.txt'),
                'test': {'prun_sl': (-1, 0.6451612903225806),
                         'prun_sr': (-1, 0.5507246376811594)},
            },
            {
                'ensemble_number': 4,
                'interfaces': [0.55, 0.69, 0.75],
                'file': os.path.join(pptisdir, 'pathensemble4.txt'),
                'test': {'prun_sl': (-1, 0.4807692307692308),
                         'prun_sr': (-1, 0.723404255319149)},
            },
            {
                'ensemble_number': 5,
                'interfaces': [0.69, 0.75, 0.9],
                'file': os.path.join(pptisdir, 'pathensemble5.txt'),
                'test': {'prun_sl': (-1, 0.19540229885057472),
                         'prun_sr': (-1, 0.8461538461538461)},
            }
        ]
        settings = {'analysis': SECTIONS['analysis']}
        detects = [0.2, 0.325, 0.55, 0.69, 0.75, 0.9]
        p_types = []
        results = []
        for ens, dec in zip(ensembles, detects):
            filename = os.path.join(HERE, ens['file'])
            settings['tis'] = {'detect': dec}
            raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)

            res = analyse_repptis_ensemble(raw_data, settings)
            for key, val in ens['test'].items():
                self.assertAlmostEqual(val[-1], res[key][val[0]])
            if ens['ensemble_number'] > 0:
                p_types.append(res['ptypes'])
            results.append(res)
        pcrossdict = repptis_running_pcross_analysis(p_types)
        self.assertAlmostEqual(pcrossdict['overall-prun'][-1],
                               0.00010317722881789884)
        self.assertAlmostEqual(pcrossdict['overall-error'][-1],
                               0.4873103759019344)
        flux = retis_flux(results[0], results[1], 0.01)
        self.assertAlmostEqual(flux[0], 0.7346382685689098)
        self.assertAlmostEqual(flux[1], 0.19580016239096693)
        rate = retis_rate(pcrossdict['overall-prun'][-1],
                          pcrossdict['overall-error'][-1],
                          flux[0], flux[1])
        self.assertAlmostEqual(rate[0], 7.579794073451942e-05)
        self.assertAlmostEqual(rate[1], 0.5251753098290262)

    def test_path_analysis_pptis_skiplines(self):
        """Test skiplines path ensemble analysis for REPPTIS simulations."""
        # we take the same pathensemble files as the previous test, but skip
        # 3 of the 100 lines.
        pptisdir = "pptis"
        ensembles = [
            {
                'ensemble_number': 0,
                'interfaces': [0.1, 0.2, 0.2],
                'file': os.path.join(pptisdir, 'pathensemble0.txt'),
                'test': {'fluxlength': (0, 91.21649484536083)},
            },
            {
                'ensemble_number': 1,
                'interfaces': [0.2, 0.2, 0.325],
                'file': os.path.join(pptisdir, 'pathensemble1.txt'),
                'test': {'prun_sl': (-1, 0.07692307692307693),
                         'prun_sr': (-1, 1.)},
            },
            {
                'ensemble_number': 2,
                'interfaces': [0.2, 0.325, 0.55],
                'file': os.path.join(pptisdir, 'pathensemble2.txt'),
                'test': {'prun_sl': (-1, 0.01098901098901099),
                         'prun_sr': (-1, 1.)},
            },
            {
                'ensemble_number': 3,
                'interfaces': [0.325, 0.55, 0.69],
                'file': os.path.join(pptisdir, 'pathensemble3.txt'),
                'test': {'prun_sl': (-1, 0.6333333333333333),
                         'prun_sr': (-1, 0.5757575757575758)},
            },
            {
                'ensemble_number': 4,
                'interfaces': [0.55, 0.69, 0.75],
                'file': os.path.join(pptisdir, 'pathensemble4.txt'),
                'test': {'prun_sl': (-1, 0.47058823529411764),
                         'prun_sr': (-1, 0.717391304347826)},
            },
            {
                'ensemble_number': 5,
                'interfaces': [0.69, 0.75, 0.9],
                'file': os.path.join(pptisdir, 'pathensemble5.txt'),
                'test': {'prun_sl': (-1, 0.21518987341772153),
                         'prun_sr': (-1, 0.8333333333333334)},
            }
        ]
        settings = {'analysis': SECTIONS['analysis'].copy()}
        settings['analysis']['skip'] = 3
        detects = [0.2, 0.325, 0.55, 0.69, 0.75, 0.9]
        p_types = []
        results = []
        for ens, dec in zip(ensembles, detects):
            filename = os.path.join(HERE, ens['file'])
            settings['tis'] = {'detect': dec}
            raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
            res = analyse_repptis_ensemble(raw_data, settings)
            for key, val in ens['test'].items():
                self.assertAlmostEqual(val[-1], res[key][val[0]])
            if ens['ensemble_number'] > 0:
                p_types.append(res['ptypes'])
            results.append(res)
        pcrossdict = repptis_running_pcross_analysis(p_types)
        self.assertAlmostEqual(pcrossdict['overall-prun'][-1],
                               0.00011041092698379775)
        self.assertAlmostEqual(pcrossdict['overall-error'][-1],
                               0.5460014951716033)
        flux = retis_flux(results[0], results[1], 0.01)
        self.assertAlmostEqual(flux[0], 0.7391039317281317)
        self.assertAlmostEqual(flux[1], 0.21291472586483318)
        rate = retis_rate(pcrossdict['overall-prun'][-1],
                          pcrossdict['overall-error'][-1],
                          flux[0], flux[1])
        self.assertAlmostEqual(rate[0], 8.160515023947258e-05)
        self.assertAlmostEqual(rate[1], 0.5860463405053592)

    def test_path_analysis_pptis_local_nans(self):
        """Test REPPTIS analysis when some local probs are NaN.

        We still want a PDF report to be written about the other
        ensembles, so we make sure that NaNs are outputted without
        a crash occurring.

        """
        pptisdir = "pptis"
        ensembles = [
            {
                'ensemble_number': 0,
                'interfaces': [0.1, 0.2, 0.2],
                'file': os.path.join(pptisdir, 'pathensemble0.txt'),
                'test': {'fluxlength': (0, 91.21649484536083)},
            },
            {
                'ensemble_number': 1,
                'interfaces': [0.2, 0.2, 0.325],
                'file': os.path.join(pptisdir, 'pathensemble1.txt'),
                'test': {'prun_sl': (-1, 0.07692307692307693),
                         'prun_sr': (-1, 1.)},
            },
            {
                'ensemble_number': 2,
                'interfaces': [0.2, 0.325, 0.55],
                'file': os.path.join(pptisdir, 'pathensemble2_noRacc.txt'),
                'test': {'prun_sl': (-1, 0.010309278350515464),
                         'prun_sr': (-1, 0.0)},
            },
            {
                'ensemble_number': 3,
                'interfaces': [0.325, 0.55, 0.69],
                'file': os.path.join(pptisdir, 'pathensemble3_noLacc.txt'),
                'test': {'prun_sl': (-1, 0.0),
                         'prun_sr': (-1, 0.4742268041237113)},
            },
            {
                'ensemble_number': 4,
                'interfaces': [0.55, 0.69, 0.75],
                'file': os.path.join(pptisdir, 'pathensemble4.txt'),
                'test': {'prun_sl': (-1, 0.47058823529411764),
                         'prun_sr': (-1, 0.717391304347826)},
            },
            {
                'ensemble_number': 5,
                'interfaces': [0.69, 0.75, 0.9],
                'file': os.path.join(pptisdir, 'pathensemble5.txt'),
                'test': {'prun_sl': (-1, 0.21518987341772153),
                         'prun_sr': (-1, 0.8333333333333334)},
            }
        ]
        settings = {'analysis': SECTIONS['analysis'].copy()}
        settings['analysis']['skip'] = 3
        detects = [0.2, 0.325, 0.55, 0.69, 0.75, 0.9]
        p_types = []
        results = []
        for ens, dec in zip(ensembles, detects):
            filename = os.path.join(HERE, ens['file'])
            settings['tis'] = {'detect': dec}
            raw_data = PathEnsembleFile(filename, 'r', ensemble_settings=ens)
            res = analyse_repptis_ensemble(raw_data, settings)
            for key, val in ens['test'].items():
                self.assertAlmostEqual(val[-1], res[key][val[0]])
            if ens['ensemble_number'] > 0:
                p_types.append(res['ptypes'])
            results.append(res)
        pcrossdict = repptis_running_pcross_analysis(p_types)
        self.assertTrue(np.isnan(pcrossdict['overall-prun'][-1]))
        self.assertTrue(np.isnan(pcrossdict['overall-error'][-1]))
        flux = retis_flux(results[0], results[1], 0.01)
        self.assertAlmostEqual(flux[0], 0.7391039317281317)
        self.assertAlmostEqual(flux[1], 0.21291472586483318)
        rate = retis_rate(pcrossdict['overall-prun'][-1],
                          pcrossdict['overall-error'][-1],
                          flux[0], flux[1])
        self.assertTrue(np.isnan(rate[0]))
        self.assertTrue(np.isnan(rate[1]))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the ForceField and PotentialFunction classes."""
import logging
import unittest
import numpy as np
from pyretis.core.system import System
from pyretis.core.particles import Particles
from pyretis.forcefield import ForceField, PotentialFunction
logging.disable(logging.CRITICAL)


class TestPotential(PotentialFunction):
    """A potential function to use in tests."""

    def __init__(self, desc='Test potential'):
        super().__init__(dim=1, desc=desc)
        self.params = {'a': 10}

    def potential(self, system):
        """Evaluate the potential."""
        pos = system.particles.pos
        return np.sum(pos * pos * self.params['a'])

    def force(self, system):
        """Evaluate force and virial."""
        pos = system.particles.pos
        return pos * self.params['a'], 2.0 * pos

    def potential_and_force(self, system):
        """Evaluate potential, force and virial."""
        pot = self.potential(system)
        force, virial = self.force(system)
        return pot, force, virial


class TestForceField(unittest.TestCase):
    """Test set-up of force fields."""

    def test_forcefield_class(self):
        """Test functionality of the ForceField class."""
        system = System()
        system.particles = Particles(dim=system.get_dim())
        system.add_particle(1.0)
        forcefield = ForceField('Generic testing force field')
        param1 = {'a': 1.0}
        pot1 = TestPotential()
        forcefield.add_potential(pot1, parameters=param1)

        force, virial = forcefield.evaluate_force(system)
        self.assertAlmostEqual(1.0, force)
        self.assertAlmostEqual(2.0, virial)

        vpot = forcefield.evaluate_potential(system)
        self.assertAlmostEqual(1.0, vpot)

        vpot, force, virial = forcefield.evaluate_potential_and_force(system)
        self.assertAlmostEqual(1.0, force)
        self.assertAlmostEqual(2.0, virial)
        self.assertAlmostEqual(1.0, vpot)

        param2 = {'a': 2.0}
        forcefield.update_potential_parameters(pot1, param2)

        vpot, force, virial = forcefield.evaluate_potential_and_force(system)
        self.assertAlmostEqual(2.0, force)
        self.assertAlmostEqual(2.0, virial)
        self.assertAlmostEqual(2.0, vpot)

        potr, paramr = forcefield.remove_potential(pot1)
        self.assertIs(pot1, potr)
        self.assertIs(param2, paramr)

        pot2 = TestPotential()
        potr, paramr = forcefield.remove_potential(pot2)
        self.assertTrue(potr is None)
        self.assertTrue(paramr is None)
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.forcefield.forcefield',
                             level='WARNING'):
            forcefield.update_potential_parameters(pot2, param2)
        logging.disable(logging.CRITICAL)

    def test_add_potential(self):
        """Test the forcefield add potential method in more detail."""
        forcefield = ForceField('Generic testing force field')
        param1 = {'a': 1.0}
        pot1 = TestPotential()
        ret = forcefield.add_potential(pot1, parameters=param1)
        self.assertTrue(ret)
        ret = forcefield.add_potential(None, parameters=param1)
        self.assertFalse(ret)
        pot2 = TestPotential()
        forcefield = ForceField('Generic testing force field',
                                potential=[pot1, pot2],
                                params=[param1])
        self.assertTrue(pot1 in forcefield.potential)
        self.assertTrue(pot2 in forcefield.potential)
        self.assertEqual(param1, forcefield.params[0])
        self.assertTrue(forcefield.params[1] is None)

    def test_evaluation(self):
        """Test evaluation of the force field."""
        system = System()
        system.particles = Particles(dim=system.get_dim())
        system.add_particle(1.0)
        param1 = {'a': 1.0}
        pot1 = TestPotential()
        param2 = {'a': 2.0}
        pot2 = TestPotential()
        forcefield = ForceField('Generic testing force field',
                                potential=[pot1, pot2],
                                params=[param1, param2])
        vpot = forcefield.evaluate_potential(system)
        self.assertAlmostEqual(vpot, 3.0)
        _, force, virial = forcefield.evaluate_potential_and_force(system)
        self.assertTrue(np.allclose(force, np.array([[3.0]])))
        self.assertTrue(np.allclose(virial, np.array([[4.0]])))
        force, virial = forcefield.evaluate_force(system)
        self.assertTrue(np.allclose(force, np.array([[3.0]])))
        self.assertTrue(np.allclose(virial, np.array([[4.0]])))

    def test_print_potentials(self):
        """Test the printing of force field information."""
        param1 = {'a': 1.0}
        pot1 = TestPotential()
        param2 = {'a': 2.0}
        pot2 = TestPotential()
        forcefield = ForceField('Generic testing force field',
                                potential=[pot1, pot2],
                                params=[param1, param2])
        txt = forcefield.print_potentials()
        correct_txt = ['Force field: Generic testing force field',
                       '1: Test potential', '2: Test potential']
        for i, j in zip(txt.split('\n'), correct_txt):
            self.assertEqual(i.strip(), j.strip())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Merge the output from two PyRETIS simulations.

Here we merge the output of two consecutive simulations (i.e. one
is the continuation of the other). Note that we make some specific
assumtions on how the output is written and that we do not attempt
to merge trajectory archive files.

"""
import os
import sys
from pyretis.inout.common import make_dirs
from pyretis.inout.settings import parse_settings_file
from pyretis.core.pathensemble import generate_ensemble_name


def merge_files(file1, file2, outputfile):
    """Merge two files to a single file.

    This method is used to merge PyRETIS output files into a
    single file. Here we assume that the data for the first
    cycle in the second file is identical to the data for the
    last cycle in the first file, and we write this cycle only
    once to the merged output file.

    Parameters
    ----------
    file1 : string
        Path to the first file.
    file2 : string
        Path to the second file.
    outputfile : string
        The path to the merged file we will create here.

    """
    with open(outputfile, 'w') as output:
        # Write full contents of file1:
        with open(file1, 'r') as infile1:
            output.write(infile1.read())
        # Skip data for the first cycle of file2:
        skip = 1
        with open(file2, 'r') as infile2:
            for lines in infile2:
                # The pathensemble.txt files are different from the rest,
                # and we only need to skip the first line.
                if os.path.basename(file2) == 'pathensemble.txt':
                    skip -= 1
                elif lines.startswith('# Cycle'):
                    # For the other files we need to skip to the next
                    # cycle output:
                    skip -= 1
                if skip < 0:
                    output.write(lines)


def main(indir1, indir2, outdir):
    """Merge files in the two given directories."""
    settings = parse_settings_file(os.path.join(indir1, 'retis.rst'))
    inter = settings['simulation']['interfaces']
    for i, _ in enumerate(inter):
        ensemble_dir = generate_ensemble_name(i)
        output_ensemble_dir = os.path.join(outdir, ensemble_dir)
        make_dirs(output_ensemble_dir)
        for filei in {'energy.txt', 'order.txt', 'pathensemble.txt'}:
            file1 = os.path.join(indir1, ensemble_dir, filei)
            file2 = os.path.join(indir2, ensemble_dir, filei)
            merged = os.path.join(output_ensemble_dir, filei)
            merge_files(file1, file2, merged)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1], sys.argv[2], sys.argv[3]))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module defining a table-like output format.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TxtTableFormatter (:py:class:`.TxtTableFormatter`)
    A class for creating a table-like format.

PathTableFormatter (:py:class:`.PathTableFormatter`)
    A class for table-like output from path simulations.

ThermoTableFormatter (:py:class:`.ThermoTableFormatter`)
    A class for thermodynamic (energy) output. Useful for output from
    MD-simulations.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

txt_save_columns (:py:class:`.txt_save_columns`)
    For writing simple column-based output.

"""
import logging
import numpy as np
from pyretis.inout.common import create_backup
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.core.path import _GENERATED_SHORT
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

np.set_printoptions(legacy='1.21')

__all__ = [
    'TxtTableFormatter',
    'PathTableFormatter',
    'ThermoTableFormatter',
    'RETISResultFormatter',
    'txt_save_columns',
]


def txt_save_columns(outputfile, header, variables, backup=False):
    """Save variables to a text file using ``numpy.savetxt``.

    Note that the variables are assumed to be numpy.arrays of equal
    shape and that the output file may also be a compressed file in
    gzip format (this is selected by letting the output file name
    end with '.gz').

    Parameters
    ----------
    outputfile : string
        Name of the output file to create.
    header : string
        A string that will be written at the beginning of the file.
    variables : tuple or list of numpy.arrays
        These are the variables that will be saved to the text file.
    backup : boolean, optional
        Determines if we should backup old files or not.

    """
    if backup:
        msg = create_backup(outputfile)
        if msg:
            logger.warning(msg)
    nvar = len(variables)
    mat = np.zeros((len(variables[0]), nvar))
    for i, vari in enumerate(variables):
        try:
            mat[:, i] = vari
        except ValueError:
            msg = 'Could not align variables, skipping (writing zeros)'
            logger.warning(msg)
    np.savetxt(outputfile, mat, header=header)


def _fill_list(the_list, length, fillvalue=None):
    """Fill a list to a specified length.

    Parameters
    ----------
    the_list : list
        The list to fill.
    length : int
        The required length.
    fillvalue : optional
        The value to insert. If None is given the last item in the list
        is repeated.

    """
    if fillvalue is None:
        fillvalue = the_list[-1]
    while len(the_list) < length:
        the_list.append(fillvalue)


class TxtTableFormatter(OutputFormatter):
    """A class for generating table output.

    This class handles formatting of output data to a table-like
    format.

    Attributes
    ----------
    variables : list of strings
        These are the variables we will use in the table.
    fmt : string
        The format to apply to the columns.
    row_fmt : list of strings
        A list of strings used for formatting, used to construct `fmt`.
    title : string
        A title for the table.

    Example
    -------
    For creating a new table, a dictionary is convenient for
    grouping the settings:

    >>> tabl = {
    ...    'title': 'Energy output',
    ...    'var': ['step', 'temp', 'vpot'
    ...            'ekin', 'etot', 'press'],
    ...    'format': {'labels': ['Step', 'Temp', 'Pot',
    ...               'Kin', 'Tot', 'Press'],
    ...               'width': (10, 12),
    ...               'spacing': 2,
    ...               'row_fmt': ['{:> 10d}', '{:> 12.6g}']
    ...              }
    ...    }
    >>> table = TxtTableFormatter(tabl['var'], tabl['title'], **tabl['format'])


    """

    def __init__(self, variables, title, **kwargs):
        """Initialise the TxtTable object.

        Parameters
        ----------
        variables : list of strings
            These are the variables we will use in the table. If the
            header is not specified, then we will create one using
            these variables.
        title : string
            A title for the table.
        kwargs : dict
            Additional settings for the formatter. This may contain:

            * width : list of ints
                The (maximum) width of the columns. If the number of
                items in this list is smaller than the number of
                variables, we simply repeat the last width for the
                number of times we need.
            * labels : list of strings
                Table headers to use for the columns.
            * spacing : integer
                The separation between columns. The default value is 1.
            * row_fmt : list of strings
                The format to apply to the columns. If the number of
                items in this list is smaller than the number of
                variables, we simply repeat the last width for the
                number of times we need.

        """
        spacing = kwargs.get('spacing', 1)
        header = {'spacing': spacing,
                  'labels': kwargs.get('labels', list(variables))}
        width = kwargs.get('width', None)
        if width is None:
            header['width'] = [max(12, len(i)) for i in header['labels']]
        else:
            header['width'] = list(width)

        _fill_list(header['width'], len(header['labels']))

        super().__init__('TxtTableFormatter', header=header)
        self.title = title
        self.variables = variables
        row_fmt = kwargs.get('row_fmt', None)
        if row_fmt is None:
            self.row_fmt = []
            for wid in header['width']:
                if wid - 6 <= 0:
                    self.row_fmt.append(f'{{:> {wid}}}')
                else:
                    self.row_fmt.append(f'{{:> {wid}.{wid-6}g}}')
        else:
            self.row_fmt = row_fmt
        _fill_list(self.row_fmt, len(self.variables))
        str_white = ' ' * spacing
        self.fmt = str_white.join(self.row_fmt)

    def format(self, step, data):
        """Generate output from a dictionary using the requested variables.

        Parameters
        ----------
        step : int
            This is the current step number or a cycle number in a
            simulation.
        data : dict
            This is assumed to a dictionary containing the row items we
            want to format.

        Yields
        ------
        out : string
            A line with the formatted output.

        """
        var = []
        for i in self.variables:
            if i == 'step':
                var.append(step)
            else:
                var.append(data.get(i, float('nan')))
        txt = self.fmt.format(*var)
        yield txt

    def __str__(self):
        """Return a string with some info about the TxtTableFormatter."""
        msg = [f'TxtTableFormatter: "{self.title}"']
        msg += [f'\t* Variables: {self.variables}']
        msg += [f'\t* Format: {self.fmt}']
        return '\n'.join(msg)


class PathTableFormatter(TxtTableFormatter):
    """A special table output class for path ensembles.

    This object will return a table of text with a header and with
    formatted rows for a path ensemble. The table rows will contain
    data from the `PathEnsemble.nstats` variable. This table is just
    meant as output to the screen during a path ensemble simulation.
    """

    def __init__(self):
        """Initialise the PathTableFormatter."""
        title = 'Path Ensemble Statistics'
        var = ['step', 'ACC', 'BWI',
               'BTL', 'FTL', 'BTX', 'FTX']
        table_format = {'labels': ['Cycle', 'Accepted', 'BWI', 'BTL', 'FTL',
                                   'BTX', 'FTX'],
                        'width': (10, 12),
                        'spacing': 2,
                        'row_fmt': ['{:> 10d}', '{:> 12d}']}
        super().__init__(var, title, **table_format)

    def format(self, step, data):
        """Generate the output for the path table.

        Here we overload the :py:meth:`.TxtTableFormatter.format` method
        in order to write path ensemble statistics to (presumably)
        the screen.

        Parameters
        ----------
        step : int
            This is the current step number or a cycle number in a
            simulation.
        data : object like :py:class:`.PathEnsemble`
            This is the path ensemble we are generating output for.

        Yield
        -----
        out : string
            The formatted output.

        """
        row = {}
        for key in self.variables:
            if key == 'step':
                value = step
            else:
                value = data.nstats.get(key, 0)
            row[key] = value
        var = [row.get(i, float('nan')) for i in self.variables]
        yield self.fmt.format(*var)


class ThermoTableFormatter(TxtTableFormatter):
    """A special text table for energy output.

    This object will return a table of text with a header and with
    formatted rows for energy output. Typical use is in MD simulation
    where we want to display energies at different steps in the
    simulations.
    """

    def __init__(self):
        """Initialise the ThermoTableFormatter."""
        title = 'Energy Output'
        var = ['step', 'temp', 'vpot', 'ekin', 'etot', 'press']
        table_format = {'labels': ['Step', 'Temp', 'Pot',
                                   'Kin', 'Tot', 'Press'],
                        'width': (10, 12),
                        'spacing': 2,
                        'row_fmt': ['{:> 10d}', '{:> 12.6g}']}
        super().__init__(var, title, **table_format)


class RETISResultFormatter(TxtTableFormatter):
    """A special table output class for path ensembles in RETIS simulations.

    This object will return a table of text with a header and with
    formatted rows for a path ensemble. The table rows will contain
    data from the `PathEnsemble.nstats` variable. This table is just
    meant as output to the screen during a path ensemble simulation.
    """

    def __init__(self):
        """Initialise the PathTableFormatter."""
        title = 'Path Ensemble Statistics'
        var = ['pathensemble', 'step', 'ACC', 'BWI',
               'BTL', 'FTL', 'BTX', 'FTX']
        table_format = {
            'labels': [
                'Ensemble', 'Cycle', 'Accepted', 'BWI', 'BTL', 'FTL',
                'BTX', 'FTX'
            ],
            'width': (8, 8, 10),
            'spacing': 2,
            'row_fmt': ['{:>10}', '{:> 8d}', '{:> 10d}']
        }
        super().__init__(var, title, **table_format)
        self.print_header = False

    def format(self, step, data):
        """Generate the output for the path table.

        Here we overload the :py:meth:`.TxtTableFormatter.format` method
        in order to write path ensemble statistics to (presumably)
        the screen.

        Parameters
        ----------
        step : int
            This is the current step number or a cycle number in a
            simulation.
        data : object like :py:class:`.PathEnsemble`
            This is the path ensemble we are generating output for.

        Yield
        -----
        out : string
            The formatted output.

        """
        row = {}
        for key in self.variables:
            if key == 'step':
                value = step
            elif key == 'pathensemble':
                value = data.ensemble_name
            else:
                value = data.nstats.get(key, 0)
            row[key] = value
        yield (f'# Results for path ensemble {data.ensemble_name} '
               f'at cycle {step}:')
        path = data.paths[-1]
        move = _GENERATED_SHORT.get(path['generated'][0], 'unknown').lower()
        yield (f'# Generated path with status "{path["status"]}", '
               f'move "{move}" and length {path["length"]}.')
        omax = path['ordermax']
        yield f'# Order parameter max was: {omax[0]} at index {omax[1]}.'
        omin = path['ordermin']
        yield f'# Order parameter min was: {omin[0]} at index {omin[1]}.'
        yield '# Path ensemble statistics:'
        yield self.header
        var = [row.get(i, float('nan')) for i in self.variables]
        yield self.fmt.format(*var)
        yield '\n'



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test using the LAMMPS engine."""
from operator import itemgetter
import os
import colorama
import numpy as np
from pyretis.inout.common import make_dirs
from pyretis.engines.lammps import LAMMPSEngine
from pyretis.inout import print_to_screen
from pyretis.testing.helpers import clean_dir
from pyretis.testing.systemhelp import create_system_ext


HERE = os.path.abspath(os.path.dirname(__file__))
TRAJ = os.path.join(HERE, 'lammps_input', 'traj.lammpstrj')


def _convert_snapshot(snapshot):
    """Convert a LAMMPS text snapshot to numbers."""
    snapshot_new = {
        'timestep': int(snapshot['timestep'][0]),
        'number': int(snapshot['number'][0]),
    }
    box = []
    for line in snapshot['box']:
        box.append([float(i) for i in line.split()])
    snapshot_new['box'] = np.array(box)
    # Convert and sort atoms:
    atoms = []
    for line in snapshot['atoms']:
        atoms.append([float(i) for i in line.split()])
    snapshot_new['atoms'] = np.array(sorted(atoms, key=itemgetter(0)))
    return snapshot_new


def read_lammpstrj(filename):
    """Read frames in a LAMMPS trajectory."""
    snap = {}
    key = None
    data = []
    with open(filename) as infile:
        for lines in infile:
            strip_line = lines.strip()
            if strip_line.startswith('ITEM'):
                # This is the start of a new data block in a snapshot.
                # Add this data block to the current snapshot
                if key is not None:
                    snap[key] = data
                key = strip_line.split()[1].lower()
                data = []
                if key == 'timestep':
                    # This is the start of a new frame. Return the current
                    # snapshot, if any, and empty the data blocks:
                    if snap:
                        yield _convert_snapshot(snap)
                    snap = {}
                continue
            else:
                data.append(strip_line)
    if snap:
        if key is not None:
            snap[key] = data
        yield _convert_snapshot(snap)


def dump_phasepoint():
    """Use the LAMMPS engine to run a MD simulation forward in time."""
    print_to_screen('\nTesting that we can dump phase points.\n')
    engine = LAMMPSEngine('lmp_serial', 'lammps_input', 2,
                          extra_files=['dw-wca.in'])
    # Create a dummy system:
    system = create_system_ext(pos=(TRAJ, 0))
    exe_dir = os.path.join(HERE, 'dump')
    # Set up some directories:
    make_dirs(exe_dir)
    clean_dir(exe_dir)
    engine.exe_dir = exe_dir
    dumped_files = []
    for i in (0, 2, 4, 6, 8, 10):
        newpos = (TRAJ, i)
        print_to_screen(f'\t-> Dumping: {newpos}')
        system.particles.set_pos(newpos)
        engine.dump_phasepoint(system, f'dump-{i:02d}')
        pos = system.particles.get_pos()
        assert pos[1] == 0
        dumped_files.append(pos[0])
    print_to_screen('\nRunning some comparisons:\n')
    for i, frame in enumerate(read_lammpstrj(TRAJ)):
        traj = dumped_files[i]
        print_to_screen(f'\t-> Comparing for dumped frame: {traj}')
        frame_dump = [i for i in read_lammpstrj(traj)]
        assert len(frame_dump) == 1
        assert frame['number'] == frame_dump[0]['number']
        for key in ('box', 'atoms'):
            assert np.allclose(frame[key], frame_dump[0][key])
        assert frame_dump[0]['timestep'] == 0
        assert frame['timestep'] == engine.subcycles * i
        print_to_screen('\t\t->Frame was ok!', level='success')
    print_to_screen('\nConclusion: We can dump phase points.',
                    level='success')


def main():
    """Run the comparisons."""
    dump_phasepoint()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for formatting crossing data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CrossFormatter (:py:class:`.CrossFormatter`)
    A class for formatting crossing data from flux simulations.

CrossFile (:py:class:`.CrossFile`)
    A class for handling PyRETIS crossing files.

"""
import logging
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.fileio import FileIO
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = [
    'CrossFormatter',
    'CrossFile',
]


class CrossFormatter(OutputFormatter):
    """A class for formatting crossing data from flux simulations.

    This class handles formatting of crossing data. The format for
    the crossing file is three columns:

    1) The first column is the step number (an integer).

    2) The second column is the interface number (an integer). These are
       numbered from 1 (_NOT_ from 0).

    3) The direction we are moving in - `+` for the positive direction
       or `-` for the negative direction. Internally this is converted
       to an integer (`+1` or `-1`).

    """

    # Format for crossing files:
    CROSS_FMT = '{:>10d} {:>4d} {:>3s}'

    def __init__(self):
        """Initialise the crossing formatter."""
        header = {'labels': ['Step', 'Int', 'Dir'], 'width': [10, 4, 3]}
        super().__init__('CrossFormatter', header=header)

    def format(self, step, data):
        """Generate output data to be written to a file or screen.

        This will format the crossing data in a space delimited form.


        Parameters
        ----------
        step : int
            This is the current step number. It is only used here for
            debugging and can possibly be removed. However, it's useful
            to have here since this gives a common interface for all
            formatters.
        data : list of tuples
            The tuples are crossing with interfaces (if any) on the form
            `(timestep, interface, direction)` where the direction
            is '-' or '+'.

        Yields
        ------
        out : string
            The line(s) to be written.

        See Also
        --------
        :py:meth:`.check_crossing` in :py:mod:`pyretis.core.path`
        calculates the tuple `cross` which is used in this routine.

        Note
        ----
        We add 1 to the interface number here. This is for
        compatibility with the old FORTRAN code where the interfaces
        are numbered 1, 2, ... rather than 0, 1, ... .

        """
        msgtxt = f'Generating crossing data at step: {step}'
        logger.debug(msgtxt)
        for cro in data:
            if cro:
                yield self.CROSS_FMT.format(cro[0], cro[1] + 1, cro[2])

    @staticmethod
    def parse(line):
        """Parse crossing data.

        The format is described in :py:meth:`.format`, this method will
        parse this format for a _single_ row of data.

        Parameters
        ----------
        line : string
            A line to parse.

        Returns
        -------
        out[0] : integer
            The step number.
        out[1] : integer
            The interface number.
        out[2] : integer
            The direction, left (-1) or right (1).

        Note
        ----
        A '1' will be subtracted from the interface in analysis. This is
        just for backward compatibility with the old FORTRAN code.

        """
        linessplit = line.strip().split()
        step, inter = int(linessplit[0]), int(linessplit[1])
        direction = -1 if linessplit[2] == '-' else 1
        return step, inter, direction


class CrossFile(FileIO):
    """A class for handling PyRETIS crossing files."""

    def __init__(self, filename, file_mode, backup=True):
        """Create the cross file with correct formatter."""
        super().__init__(filename, file_mode, CrossFormatter(), backup=backup)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations.

Here we compare a full simulation with one where we have stopped
and restarted after 10 steps.
"""
import itertools
import os
import sys
import fileinput
import colorama
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import gridspec
from pyretis.inout import print_to_screen
from pyretis.inout.formats.snapshot import SnapshotFile
plt.style.use('seaborn-v0_8')


def remove_comments(filename):
    """Remove commented lines if present."""
    for line in fileinput.input(filename, inplace=1):
        if '#' not in line:
            sys.stdout.write(line)


def snapshot_difference(snap1, snap2):
    """Calculate difference between two snapshots."""
    xyz1 = np.column_stack((snap1['x'], snap1['y'], snap1['z']))
    xyz2 = np.column_stack((snap2['x'], snap2['y'], snap2['z']))
    diff = (xyz1 - xyz2)**2
    dsum = np.einsum('ij,ij -> i', diff, diff)
    vel1 = np.column_stack((snap1['vx'], snap1['vy'], snap1['vz']))
    vel2 = np.column_stack((snap2['vx'], snap2['vy'], snap2['vz']))
    diffv = (vel1 - vel2)**2
    dsumv = np.einsum('ij,ij -> i', diffv, diffv)
    return sum(dsum), sum(dsumv)


def compare_traj(traj11, traj12, traj2, tol=1e-12):
    """Compare two trajectories from PyRETIS.

    Here we calculate the mean squared error for the two
    trajectories.

    Parameters
    ----------
    traj11 : string
        A trajectory to open, part 1.
    traj12 : string
        A trajectory to open, part 2.
    traj2 : string
        A trajectory file to open.
    tol : float
        A tolerance for comparing numbers.

    Returns
    -------
    None, just prints out the result of the comparison.

    """
    remove_comments(traj12)
    print_to_screen('Comparing trajectories', level='info')
    print_to_screen('Checking mean squared error...')
    file11 = SnapshotFile(traj11, 'r').load()
    file12 = SnapshotFile(traj12, 'r').load()
    next(file12)  # skip the first config
    file1 = itertools.chain(file11, file12)
    file2 = SnapshotFile(traj2, 'r').load()
    error, error_v = [], []
    for snap1, snap2 in zip(file1, file2):
        pose, vele = snapshot_difference(snap1, snap2)
        error.append(pose)
        error_v.append(vele)
    if next(file1, False) or next(file2, False):
        print_to_screen('Number of lines are incorrect ',
                        level='error')
        return 1
    val1 = print_error_assessment(np.mean(error), 'positions', tol)
    val2 = print_error_assessment(np.mean(error_v), 'velocities', tol)
    return val1 + val2


def compare_step_output(file11, file12, file2):
    """Compare step-wise output from PyRETIS.

    Here, we assume that the output in file2 is obtained
    by concatenating file11 and file12. file11 and file12
    should have one line in common: the final line in file11
    should be identical to the first line in file12.

    Parameters
    ----------
    file11 : string
        A file to open, part 1.
    file12 : string
        A file to open, part 2.
    file2 : string
        A file to open, we will check if this one is
        equal (line-by-line) to file11+file12.

    Returns
    -------
    None, just prints out the result of the comparison.

    """
    remove_comments(file12)
    print_to_screen('\nComparing files:', level='info')
    names = [os.path.split(i)[1] for i in (file11, file12, file2)]
    print_to_screen('{} + {} == {}?'.format(*names))
    file11_h = open(file11, 'r')
    file12_h = open(file12, 'r')
    file2_h = open(file2, 'r')
    first_line = next(file12_h)
    error = False
    for i, (data1, data2) in enumerate(zip(itertools.chain(file11_h, file12_h),
                                           file2_h)):
        if not data1 == data2:
            print_to_screen(f'Error for line no: {i}', level='error')
            print_to_screen('Lines were:', level='error')
            print_to_screen(data1.strip(), level='error')
            print_to_screen(data2.strip(), level='error')
            error = True
            break
    file11_h.close()
    file12_h.close()
    file2_h.close()
    if not error:
        print_to_screen('Joined files contain same data!')
        last_line, len11 = read_last_line(file11)
        if last_line == first_line:
            print_to_screen('First in {} = last in {}'.format(*names[:2]))
            _, len12 = read_last_line(file12)
            _, len2 = read_last_line(file2)
            if len11 + len12 == len2 + 1:
                print_to_screen('Number of lines are correct.')
                print_to_screen('Files are equal!', level='success')
                return 0
            print_to_screen(('Number of lines are incorrect ',
                             f'{len11 + len12} != {len2 + 1}'),
                            level='error')
            return 1
        print_to_screen('First in {} != last in {}'.format(*names[:2]),
                        level='error')
        return 1
    return 1


def read_last_line(filename):
    """Read the last line from a file + count number of lines in the file."""
    i = 0
    last_line = None
    with open(filename, 'r') as infile:
        for lines in infile:
            last_line = lines
            i += 1
    return last_line, i


def print_error_assessment(error, what, tol):
    """Print out some error info."""
    if abs(error) < tol:
        lev = 'success'
        val = 0
    else:
        lev = 'error'
        val = 1
    print_to_screen(f'Mean error - {what}: {error}',
                    level=lev)
    return val


def make_fig():
    """Plot for comparison."""
    fig1 = plt.figure(figsize=(12, 6))
    grid = gridspec.GridSpec(2, 2)
    ax1 = fig1.add_subplot(grid[:, 0])
    ax1.plot([], [], label='Potential', lw=0, alpha=0)
    ax1.plot([], [], label='Kinetic', lw=0, alpha=0)
    ax1.plot([], [], label='Total', lw=0, alpha=0)
    ax1.set_xlabel('Step no.')
    ax1.set_ylabel('Energy per particle')
    ax2 = fig1.add_subplot(grid[0, 1])
    ax2.set_ylabel('Temperature')
    ax3 = fig1.add_subplot(grid[1, 1])
    ax3.set_xlabel('Step no.')
    ax3.set_ylabel('Pressure')
    axes = (ax1, ax2, ax3)
    return fig1, axes


def plot_in_ax(axes, infile, lab, fat=False, colors=None, style='-'):
    """Just do some plotting."""
    ax1, ax2, ax3 = axes
    data = np.loadtxt(infile)
    if fat:
        width = 7
    else:
        width = 3
    lines = []
    for i, idx in enumerate((2, 3, 4)):
        if colors is None:
            line, = ax1.plot(data[:, 0], data[:, idx], label=lab,
                             ls=style, lw=width, alpha=0.8)
        else:
            line, = ax1.plot(data[:, 0], data[:, idx], label=lab,
                             ls=style, lw=width, alpha=0.8, color=colors[i])
        lines.append(line)
    ax2.plot(data[:, 0], data[:, 1], label=lab, ls=style, lw=width, alpha=0.9)
    ax3.plot(data[:, 0], data[:, 5], label=lab, ls=style, lw=width, alpha=0.9)
    return lines


def make_plots():
    """Just plot some energies for comparison."""
    figure, axes = make_fig()

    plot_in_ax(
        axes,
        os.path.join('run-full', 'md-full-thermo.txt'),
        'full',
        fat=True,
        style='-'
    )
    lines = plot_in_ax(
        axes,
        os.path.join('run-10', 'md-10-thermo.txt'),
        'restart-part1',
        style='--'
    )
    colors = [i.get_color() for i in lines]
    plot_in_ax(
        axes,
        os.path.join('run-10-100', 'md-10-100-thermo.txt'),
        'restart-part2',
        style=':', colors=colors)
    axes[0].legend(prop={'size': 'medium'}, ncol=4)
    axes[1].legend(prop={'size': 'medium'})
    axes[2].legend(prop={'size': 'medium'})
    figure.subplots_adjust(bottom=0.12, right=0.95, top=0.95,
                           left=0.08, wspace=0.2)
    return figure


def main(args):
    """Run the comparison."""
    val1 = compare_traj(
        os.path.join('run-10', 'md-10-traj.xyz'),
        os.path.join('run-10-100', 'md-10-100-traj.xyz'),
        os.path.join('run-full', 'md-full-traj.xyz'),
        tol=1e-12
    )
    val2 = compare_step_output(
        os.path.join('run-10', 'md-10-thermo.txt'),
        os.path.join('run-10-100', 'md-10-100-thermo.txt'),
        os.path.join('run-full', 'md-full-thermo.txt'),
    )
    val3 = compare_step_output(
        os.path.join('run-10', 'md-10-energy.txt'),
        os.path.join('run-10-100', 'md-10-100-energy.txt'),
        os.path.join('run-full', 'md-full-energy.txt'),
    )
    if 'make_plot' in args:
        fig = make_plots()
        fig.savefig('compare.png')
        plt.show()
    return val1 + val2 + val3


if __name__ == '__main__':
    colorama.init(autoreset=True)
    sys.exit(main(sys.argv[1:]))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package defines different simulations for use with PyRETIS.

The different simulations are defined as objects which inherit from
the base Simulation object defined in `simulation.py`. The simulation
object defines as simulation as a series of tasks to be executed,
typically at each step of the simulation. These tasks may produce
results which can be outputted to the user in some way.

Package structure
-----------------

Modules
~~~~~~~

md_simulation.py (:py:mod:`pyretis.simulation.md_simulation`)
    Defines simulation classes for molecular dynamics simulations.

mc_simulation.py (:py:mod:`pyretis.simulation.mc_simulation`)
    Define simulation classes for Monte Carlo simulations.

path_simulation.py (:py:mod:`pyretis.simulation.path_simulation`)
    Defines simulation classes for path simulations.

simulation.py (:py:mod:`pyretis.simulation.simulation`)
    Defines the Simulation class which is the base class for
    simulations.

simulation_task.py (:py:mod:`pyretis.simulation.simulation_task`)
    Defines classes for the handling of simulation tasks.

Important classes defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Simulation (:py:class:`.Simulation`)
    The base class for simulations.

SimulationTask (:py:class:`.SimulationTask`)
    A class for creating tasks for simulations.

SimulationSingleTIS (:py:class:`.SimulationSingleTIS`)
    A class for running a TIS simulation for a single ensemble.

SimulationRETIS (:py:class:`.SimulationRETIS`)
    A class for running a RETIS simulation for a set of ensembles.
"""
from .simulation import Simulation
from .simulation_task import SimulationTask
from .mc_simulation import UmbrellaWindowSimulation
from .md_simulation import SimulationMD, SimulationNVE, SimulationMDFlux
from .path_simulation import SimulationTIS, SimulationRETIS
